import numpy as np
from numpy import sin, cos, tan, arccos, arcsin, arctan, sqrt, exp, log, pi
from numpy import linalg as la
import csv
import sys
from pprint import pprint
import matplotlib.pyplot as plt


from tklib.tkobject import tkObject
from tklib.tkcrystal.tkcrystalobject import tkCrystalObject
from tklib.tkcrystal.tkspacegroup import tkSpaceGroup
from tklib.tkcrystal.tkatomtype import tkAtomType
from tklib.tkcrystal.tkatomsite import tkAtomSite
from tklib.tksci.tksci import a0, NA, pi, pi2, torad, todeg, round, Factorize, IsFactor, Factors
from tklib.tksci.tksci import degcos, degsin, degtan, acos, asin, atan, degacos, degasin, degatan
from tklib.tkutils import SplitFilePath, colors
import tklib.tkre
from tklib.tkre import DelQuote



"""
Package to manage crystal structure
 class tkCrystal
"""


class tkCrystal(tkCrystalObject):
    def __init__(self, **args):
        super(tkCrystal, self).__init__(**args)
        self.Initialize()
        self.update(**args)

    def __del__(self):
        super(tkCrystal, self).__del__()

    def __str__(self):
        return self.ClassPath()


    def Initialize(self):
        pass

    def PrintInf(self, key = None):
        cry = self
        if key is None or "header" in key:
            print("")
            print("====== tkcrystal.Print start ======") 
            try:
                print("CIF path = ", cry.path)
            except:
                pass

        if key is None or "sample" in key:
            print("CrystalName = ", cry.CrystalName(Create = 1))
            print("SampleName = ",  cry.SampleName(Create = 1))
#            print("CrystalName = ", cry.__CrystalName)
#           print("SampleName = ",  cry.__SampleName)

        if key is None or "symmetry" in key:
            print("")
            print("Symmetry:")
            ls = cry.LatticeSystem()
            print("Lattice system: ", ls)
            la = cry.LatticeAxis()
            print("Lattice axis: ", la)
            SPGName, iSPG, iSet = cry.GetSpaceGroupInf()
            print("Space group: [{}]  #{}  set {}".format(SPGName, iSPG, iSet))
            for i in range(self.nSymmetryOperation()):
                sym = self.SymmetryOperation(i)
                print("  {}: {}".format(i,sym))

        if key is None or "cell" in key:
            print("")
            print("Real space:")
            a, b, c, alpha, beta, gamma = cry.LatticeParameters()
            print("cell: {:12.8f} {:12.8f} {:12.8f} A   {:10.6f} {:10.6f} {:10.6f}".format(a, b, c, alpha, beta, gamma))

        if key is None or "lattice" in key:
            gij, Rgij = cry.Metrics()
            print("Metrics gij:")
            pprint(gij)
            aij = cry.LatticeVectors()
            print("Lattice vectors aij:")
            pprint(aij)
#        cry.CalculateVolume()
            V = cry.Volume()
            print("Volume: {} A^3".format(V))

        if key is None or "reciprocal" in key:
            print("")
            print("Reciprocal space:")
            ra, rb, rc, ralpha, rbeta, rgamm = cry.ReciprocalLatticeParameters()
            print("cell: {:12.6f} {:12.6f} {:12.6f} A^-1   {} {} {}".format(ra, rb, rc, ralpha, rbeta, rgamm))
            print("Metrics Rgij (A):")
            pprint(Rgij)
            Raij = cry.ReciprocalLatticeVectors()
            print("Lattice vectors Raij (A^-1):")
            pprint(Raij)
            RV = cry.ReciprocalVolume()
            print("Volume: {} A^-3".format(RV))

        if key is None or "properties" in key:
            print("")
            print("Properties:")
            cc    = cry.ChemicalComposition()
            sumcc = cry.SumChemicalComposition()
            fu    = cry.FormulaUnit()
            MW    = round(cry.UnitcellWeight(), 6)
            d     = cry.Density()
            ad    = cry.AtomDensity()
            print("  Chemical composition: ", cc)
            print("  Sum chemical composition: ", sumcc)
            print("  Formula unit: ", fu)
            print("  Unit cell weight: ", MW)
            print("  Density: ", d, "g/cm3")
            print("  Atom density: {:12.6e} /cm3".format(ad))

        if key is None or "atomtypes" in key:
            print("")
            print("Atom types:")
            AtomTypes = self.AtomTypeList()
            for t in AtomTypes:
                typea = t.AtomType()
                typeo = t.AtomTypeOnly()
                charge = t.Charge()
                AN     = t.AtomicNumber()
                M      = t.AtomicMass()
                print("  %4s (%2s) charge=%8.4f  [Z=%3d  M=%6.4f]" % (typea, typeo, charge, AN, M))

        if key is None or "asymsites" in key:
            print("")
            print("Asymmetric atom sites:")
            AtomSites = self.AtomSiteList()
            for atom in AtomSites:
                label     = atom.Label()
                atomname  = atom.AtomName()
                atomname0 = atom.AtomNameOnly()
                charge    = atom.Charge()
                pos       = atom.Position()
                occ       = atom.Occupancy()
                id        = atom.IdAsymmetricAtomSite()
                m         = atom.Multiplicity()
                print("  %3d: %5s: %4s charge=%6.3f (%6.3f, %6.3f, %6.3f) occ=%8.4f  m=%d" 
                    % (id, label, atomname, charge, pos[0], pos[1], pos[2], occ, m))

        if key is None or "allsites" in key:
            print("")
            print("Expanded atom sites:")
            AtomSites = self.ExpandedAtomSiteList()
            for atom in AtomSites:
                label     = atom.Label()
                atomname  = atom.AtomName()
                atomname0 = atom.AtomNameOnly()
                charge    = atom.Charge()
                pos       = atom.Position()
                occ       = atom.Occupancy()
                id        = atom.IdAsymmetricAtomSite()
                m         = atom.Multiplicity()
                print("  %3d: %5s: %4s charge=%6.3f (%6.3f, %6.3f, %6.3f) occ=%8.4f  m=%d" 
                    % (id, label, atomname, charge, pos[0], pos[1], pos[2], occ, m))

        if key is None or "header" in key:
            print("====== tkcrystal.Print end ======") 

    def to_dict(self):
        crystal_name = self.CrystalName(Create = 1)
        sample_name  = self.SampleName(Create = 1)

        SPGName, iSPG, iSet = self.GetSpaceGroupInf()
        ls = self.LatticeSystem()
        la = self.LatticeAxis()
        spacegroup = { "source": "tklib.tkcrystal.tkspacegroup",
                       "symbol": SPGName,
                       "number": iSPG,
                       "set"   : iSet,
                       "point_group": "null",
                       "crystal_system": ls,
                       "hall": "null"
                      }

        a, b, c, alpha, beta, gamma = self.LatticeParameters()
        a = round(a, 8)
        b = round(a, 8)
        c = round(a, 8)
        alpha = round(a, 8)
        beta  = round(a, 8)
        gamma = round(a, 8)
        V = round(self.Volume(), 8)
        gij, Rgij = self.Metrics()
        aij = self.LatticeVectors()

        cc    = self.ChemicalComposition()
        sumcc = self.SumChemicalComposition()
        fu    = self.FormulaUnit()
        MW    = round(self.UnitcellWeight(), 6)
        d     = round(self.Density(), 8)
        ad    = round(self.AtomDensity(), 8)

        AtomTypes = self.AtomTypeList()
        elements = []
        for t in AtomTypes:
            typea = t.AtomType()
            typeo = t.AtomTypeOnly()
            charge = t.Charge()
            AN     = t.AtomicNumber()
            M      = t.AtomicMass()
            
            elements.append(typeo)

#        AtomSites = self.AtomSiteList()
        AtomSites = self.ExpandedAtomSiteList()
        total_charge = 0.0
        sites = []
        for atom in AtomSites:
            label     = atom.Label()
            atomname  = atom.AtomName()
            atomname0 = atom.AtomNameOnly()
            charge    = atom.Charge()
            pos       = atom.Position()
            pos = [round(pos[i], 8) for i in range(3)]
            xc, yc, zc = self.FractionalToCartesian(*pos)
            f         = atom.Force()
            if f[0] is None:
                f = ["null", "null", "null"]
            v         = atom.Velocity()
            if v[0] is None:
                v = ["null", "null", "null"]
            occ       = atom.Occupancy()
            id        = atom.IdAsymmetricAtomSite()
            m         = atom.Multiplicity()
             
            total_charge += charge
            species = { "element": atomname, "occu": occ, "charge": charge,
                        "xyz": [xc, yc, zc], "abc": pos, "force": f, "velocity": v, 
                        "properties": {} }
            site = { "label": label, "species": [species] }
            sites.append(site)

        lattice = {
            "matrix": aij.tolist(),
            "a": a,
            "b": b,
            "c": c,
            "alpha": alpha,
            "beta" : beta,
            "gamma": gamma,
            "volume": V,
            }

        d = {}
        d["structure"] = {
            "@module": "tklib.tkcrystal.tkcrystal",
            "@class": "Structure",
            "charge": total_charge,
            "spacegroup": spacegroup,
            "elements": elements,
            "nelements": len(elements),
            "nsites": len(AtomSites),
            "lattice": lattice,
            "sites": sites
            }

        return d

    def draw_box(self, ax, aij, nrange, linecolor = 'black'):
# (0,0,0) -> ax
        ax.plot([0.0, aij[0][0]], 
                [0.0, aij[0][1]], 
                [0.0, aij[0][2]], color = linecolor)
# (0,0,0) -> ay
        ax.plot([0.0, aij[1][0]], 
                [0.0, aij[1][1]], 
                [0.0, aij[1][2]], color = linecolor)
# (0,0,0) -> az
        ax.plot([0.0, aij[2][0]], 
                [0.0, aij[2][1]], 
                [0.0, aij[2][2]], color = linecolor)

# ax -> ax + ay
        ax.plot([aij[0][0], aij[0][0] + aij[1][0]], 
                [aij[0][1], aij[0][1] + aij[1][1]], 
                [aij[0][2], aij[0][2] + aij[1][2]], color = linecolor)
# ax -> ax + az
        ax.plot([aij[0][0], aij[0][0] + aij[2][0]], 
                [aij[0][1], aij[0][1] + aij[2][1]], 
                [aij[0][2], aij[0][2] + aij[2][2]], color = linecolor)

# ay -> ay + ax
        ax.plot([aij[1][0], aij[1][0] + aij[0][0]], 
                [aij[1][1], aij[1][1] + aij[0][1]], 
                [aij[1][2], aij[1][2] + aij[0][2]], color = linecolor)
# ay -> ay + az
        ax.plot([aij[1][0], aij[1][0] + aij[2][0]], 
                [aij[1][1], aij[1][1] + aij[2][1]], 
                [aij[1][2], aij[1][2] + aij[2][2]], color = linecolor)

# az -> az + ax
        ax.plot([aij[2][0], aij[2][0] + aij[0][0]], 
                [aij[2][1], aij[2][1] + aij[0][1]], 
                [aij[2][2], aij[2][2] + aij[0][2]], color = linecolor)
# az -> ax + ay
        ax.plot([aij[2][0], aij[2][0] + aij[1][0]], 
                [aij[2][1], aij[2][1] + aij[1][1]], 
                [aij[2][2], aij[2][2] + aij[1][2]], color = linecolor)

# ax + ay -> ax + ay + az
        ax.plot([aij[0][0] + aij[1][0], aij[0][0] + aij[1][0] + aij[2][0]], 
                [aij[0][1] + aij[1][1], aij[0][1] + aij[1][1] + aij[2][1]], 
                [aij[0][2] + aij[1][2], aij[0][2] + aij[1][2] + aij[2][2]], color = linecolor)

# ax + az -> ax + ay + az
        ax.plot([aij[0][0] + aij[2][0], aij[0][0] + aij[1][0] + aij[2][0]], 
                [aij[0][1] + aij[2][1], aij[0][1] + aij[1][1] + aij[2][1]], 
                [aij[0][2] + aij[2][2], aij[0][2] + aij[1][2] + aij[2][2]], color = linecolor)

# ay + az -> ax + ay + az
        ax.plot([aij[1][0] + aij[2][0], aij[0][0] + aij[1][0] + aij[2][0]], 
                [aij[1][1] + aij[2][1], aij[0][1] + aij[1][1] + aij[2][1]], 
                [aij[1][2] + aij[2][2], aij[0][2] + aij[1][2] + aij[2][2]], color = linecolor)

    def draw_unitcell(self, ax, nrange, kr = 1000.0, linecolor = 'black', color = 'black', atomcolor = None):
        if atomcolor is None:
            atomcolor = colors

        ncolor = len(atomcolor)

        aij = self.LatticeVectors()

        self.draw_box(ax, aij, nrange, linecolor)

        AtomTypes = self.AtomTypeList()
        atom_types = {}
        for t in AtomTypes:
#            typea = t.AtomType()
            typeo = t.AtomTypeOnly()
#            charge = t.Charge()
#            AN     = t.AtomicNumber()
#            M      = t.AtomicMass()

            atom_types[typeo] = 1

        c = 0
        for key in atom_types.keys():
            atom_types[key] = c
            c += 1

        AtomSites = self.ExpandedAtomSiteList()
        drawn_atoms = {}
        for i in range(len(AtomSites)):
            atom = AtomSites[i]
#            label     = atom.Label()
#            atomname  = atom.AtomName()
            atomname0 = atom.AtomNameOnly()
#            charge    = atom.Charge()
            pos       = atom.Position()
            xc, yc, zc = self.FractionalToCartesian(*pos)
#            occ       = atom.Occupancy()
#            id        = atom.IdAsymmetricAtomSite()
#            m         = atom.Multiplicity()

            atom_types
            ion_radius = 1.0
            
            acolor = atomcolor[atom_types[atomname0] % ncolor]

            if drawn_atoms.get(atomname0, None) is None:
                ax.scatter([xc], [yc], [zc], label = atomname0, marker = 'o', c = acolor, linewidth = 0.5, edgecolors = 'white', s = kr * ion_radius)
                drawn_atoms[atomname0] = 1
            else:
                ax.scatter([xc], [yc], [zc], marker = 'o', c = acolor, linewidth = 0.5, edgecolors = 'white', s = kr * ion_radius)

        ax.legend()

# kr: Coefficient for atomic size to plot
# rmin: Distance to judge identical atom site, in angstrom
# nrange: Range of unit cells to draw crystal structure
    def draw(self, figsize = (10, 10), nrange = [[-0.1, 1.1], [-0.1, 1.1], [-0.1, 1.1]], rmin = 0.1, kr = 100.0):
        fig = plt.figure(figsize = figsize)
        ax = fig.add_subplot(111, projection = '3d')

# Real space unit cell
        self.draw_unitcell(ax, nrange, kr, linecolor = 'black')

# Note: set_aspect() is not implemented for 3D plots
#    ax.set_aspect('equal','box')
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        zlim = ax.get_zlim()
        lim = [min([xlim[0], ylim[0], zlim[0]]), max([xlim[1], ylim[1], zlim[1]])]
        ax.set_xlim(lim)
        ax.set_ylim(lim)
        ax.set_zlim(lim)

        plt.show()


"""
package CrystalObject;
use Exporter;
@ISA = qw(Exporter);

#公開したいサブルーチン
@EXPORT = qw();

use strict;
#use Math::Matrix;
use Math::MatrixReal;
use Math::Vector::Real;

use Sci qw($pi $a0 $torad $todeg);# acos asin);
use Sci::Algorism;
use Sci::MyMatrixReal;
use Math::Complex;

use Crystal::PointGroup;
use Crystal::SpaceGroup;
use Crystal::AtomType;
use Crystal::AtomSite;
#use Crystal::Rietan;
use Crystal::RietanFP;

#==================================================================
# 静的メンバー関数
#==================================================================

sub PolarizationFactor
{
    my ($Theta, $MonochromaterAlpha) = @_;
    cos2a = 1.0;
    if(defined $MonochromaterAlpha) {
        cos2a = cos(2.0 * $MonochromaterAlpha * $torad);
        cos2a = $cos2a * $cos2a;
    }
    cos2Q = cos($torad * 2.0 * $Theta);
    return 1.0 + $cos2a * $cos2Q * $cos2Q;
}

sub LorentzFactor
{
    my ($Theta) = @_;
    cosQ  = cos($torad * $Theta);
    sinQ = sin($torad * $Theta);
#    sinQ = sin($torad * 2.0 * $Theta);
    return 1.0 / $cosQ / $sinQ / $sinQ;

#    cosQ  = cos($torad * $Theta);
    sin2Q = sin($torad * 2.0 * $Theta);
    return $cosQ / $sin2Q / $sin2Q;
}

sub TemperatureFactor
{
# Biso in Angstrom, $s in nm^-1
    my ($Biso, $s) = @_;
    $s = 0.1 * $s;
    M = $Biso * $s * $s;
    return exp(-$M);
}

#==================================================================
# 変数取得関数
#==================================================================
sub SampleName     { return shift->{SampleName}; }
sub SetSampleName  { my ($this,$s)=@_; return $this->{SampleName} = $s; }
sub SetCrystalName { my ($this,$name) = @_; return $this->{CrystalName} = $name; }
sub CrystalName    { my ($this) = @_; return $this->{CrystalName}; }
sub SPGName        { my ($this) = @_; my $SPG = $this->GetCSpaceGroup(); return $SPG->SPGName(); }
sub iSPG           { my ($this) = @_; my $SPG = $this->GetCSpaceGroup(); return $SPG->iSPG(); }
sub iSet           { my ($this) = @_; my $SPG = $this->GetCSpaceGroup(); return $SPG->iSet(); }
sub SetnAtomType              { my ($this,$n) = @_; return $this->{nAtomType} = $n; }
sub nAtomType                 { my ($this) = @_; return $this->{nAtomType}; }
sub SetnAsymmetricAtomSite    { my ($this,$n) = @_; return $this->{nAsymmetricAtomSite} = $n; }
sub nAsymmetricAtomSite       { my ($this) = @_; return $this->{nAsymmetricAtomSite}; }
sub SetnTotalExpandedAtomSite { my ($this,$n) = @_; return $this->{nTotalExpandedAtomSite} = $n; }
sub nTotalExpandedAtomSite { my ($this) = @_; return $this->{nTotalExpandedAtomSite}; }

sub GetpAtomPosition
{
    my ($this, $mode) = @_;

    pSite = $this->GetpAtomSite($mode);
    my @Pos;
    for(my $i = 0 ; $i < @$pSite ; $i++) {
        my ($x, $y, $z) = $pSite->[$i]->Position(0);
        Pos[$i][0] = $x;
        Pos[$i][1] = $y;
        Pos[$i][2] = $z;
    }
    return \@Pos;
}

sub GetpDuplicatedAtomPosition
{
    my ($this, $mode) = @_;

    my @Pos = $this->DuplicateAtomPositions($mode);
    return \@Pos;
}

sub DuplicateAtomPositions
{
    my ($this, $mode) = @_;

    pSite = $this->GetpAtomSite($mode);
    my @Pos;
    for(my $i = 0 ; $i < @$pSite ; $i++) {
        my ($x, $y, $z) = $pSite->[$i]->Position(0);
        Pos[$i][0] = $x;
        Pos[$i][1] = $y;
        Pos[$i][2] = $z;
    }
    return @Pos;
}

sub Duplicate
{
    my ($this) = @_;
    
    Crystal = new Crystal;
    $Crystal->SetSampleName($this->SampleName());
    $Crystal->SetCrystalName($this->CrystalName());

    $Crystal->SetLatticeSystem($this->LatticeSystem());
    SPGName = $this->SPGName();
    iSPG    = $this->iSPG();
    iSet    = $this->iSet();
    $Crystal->SetSpaceGroup($SPGName, $iSPG, $iSet);
    $Crystal->SetCSpaceGroup($this->GetCSpaceGroup());

    $Crystal->SetLatticeParameters($this->LatticeParameters());
    $Crystal->SetCAtomTypeList($this->GetCAtomTypeList());
    $Crystal->SetCAsymmetricAtomSiteList($this->GetCAsymmetricAtomSiteList());

    $Crystal->SetTemperature($this->Temperature());

    $Crystal->CalcMetrics();
    my @v = $this->LatticeVectors();
    if(@v > 0) {
        Crystal->SetLatticeVectors(@v);
    }
    $Crystal->ExpandCoordinates();
    
    return $Crystal;
}

sub new
{
    my ($module) = @_;
    this = {};
    bless $this;

require Math::Matrix;
require Math::MatrixReal;

    SPG = new SpaceGroup();
    $SPG->SetP1();
    $this->{"SpaceGroup"} = $SPG;

    my @AtomTypeList;
#    at1 = new AtomType();
#    push(@AtomTypeList, $at1);
#    pop(@AtomTypeList);
    $this->{"RefAtomTypeList"} = \@AtomTypeList;;
    $this->{"nAtomType"} = 0;

    my @AsAtomSiteList;
    at2 = new AtomSite();
    push(@AsAtomSiteList, $at2);
    pop(@AsAtomSiteList);
    $this->{"RefAsymmetricAtomSiteList"} = \@AsAtomSiteList;
    $this->{"nAsymmetricAtomSite"} = 0;

    my @ExAtomSiteList;
    at3 = new AtomSite();
    push(@ExAtomSiteList, $at3);
    pop(@ExAtomSiteList);
    $this->{"RefExpandedAtomSiteList"} = \@ExAtomSiteList;

    my @nAtomList;
    $this->{"RefnExpandedAtomSite"} = \@nAtomList;

    $this->{"nTotalExpandedAtomSite"}  = 0;

    return $this;
}

sub PreferentialOrientation
{
    my ($this, $IsSheet, $ho, $ko, $lo, $p1, $p2, $h, $k, $l) = @_;
    return 1.0 if($p1 == 1.0 or $p2 == 0.0);
    my @VO    = $this->GetVectorFromHKL($ho, $ko, $lo);
    my @Vhkl  = $this->GetVectorFromHKL($h, $k, $l);
    eVO   = Sci::NormalizeVector(\@VO);
    eVhkl = Sci::NormalizeVector(\@Vhkl);
    c = Sci::InnerProduct($eVO, $eVhkl);
    Q = acos(abs($c));
    $Q = $pi / 2.0 - $Q if(!$IsSheet);
#print "PO [$IsSheet]: $ho,$ko,$lo - $h,$k,$l => $Q\n";
    return $p1 + (1.0 - $p1) * exp(-$p2 * $Q * $Q);
}

sub IsVariablePosition
{
    my ($this, $x, $y, $z, $pos)= @_;

    my @x1 = ($x, $y, $z);
    c0 = new Crystal();
    $c0->SetLatticeParameters($this->LatticeParameters());
    $c0->SetCSpaceGroup($this->GetCSpaceGroup());
    $c0->AddAtomType('H');
    $c0->AddAtomSite('H1', 'H', $x1[0], $x1[1], $x1[2]);
    $c0->ExpandCoordinates();
    my @sites = $c0->GetCExpandedAtomSiteList();
    nSite = @sites;

    VariablePos = 1.0;
    $x1[$pos] += 0.14142;
    c1 = new Crystal();
    $c1->SetLatticeParameters($this->LatticeParameters());
    $c1->SetCSpaceGroup($this->GetCSpaceGroup());
    $c1->AddAtomType('H');
    $c1->AddAtomSite('H1', 'H', $x1[0], $x1[1], $x1[2]);
    $c1->ExpandCoordinates();
    @sites = $c1->GetCExpandedAtomSiteList();
    $VariablePos = 0.0 if($nSite < @sites);

    return $VariablePos;
}

sub SetMinimalLatticeParameters
{
    my ($this, @a) = @_;
    ls = lc $this->LatticeSystem();
#print "ls: $ls\n";
    if($ls eq 'cubic') {
        this->SetLatticeParameters($a[0], $a[0], $a[0], 90.0, 90.0, 90.0);
    }
    elsif($ls =~ /tetra/) {
        this->SetLatticeParameters($a[0], $a[0], $a[1], 90.0, 90.0, 90.0);
    }
    elsif($ls =~ /ortho/) {
        this->SetLatticeParameters($a[0], $a[1], $a[2], 90.0, 90.0, 90.0);
    }
    elsif($ls =~ /(rhomb|trig)/) {
        this->SetLatticeParameters($a[0], $a[0], $a[0], $a[1], $a[1], $a[1]);
    }
    elsif($ls =~ /hex/) {
        this->SetLatticeParameters($a[0], $a[0], $a[1], 90.0, 90.0, 120.0);
    }
    elsif($ls =~ /mono/) {
        this->SetLatticeParameters($a[0], $a[1], $a[2], 90.0, $a[3], 90.0);
    }
    else {
        this->SetLatticeParameters($a[0], $a[1], $a[2], $a[3], $a[4], $a[5]);
    }
    my ($a,$b,$c,$alpha,$beta,$gamma) = $this->LatticeParameters();
#print "latt: $a $b $c $alpha $beta $gamma\n";
    return ($a,$b,$c,$alpha,$beta,$gamma);
}

sub ReciprocalLatticeParameters
{
    my ($this, $UseAtomicUnit) = @_;

    Ra     = $this->{Ra};
    Rb     = $this->{Rb};
    Rc     = $this->{Rc};
    Ralpha = $this->{Ralpha};
    Rbeta  = $this->{Rbeta};
    Rgamma = $this->{Rgamma};

    k = ($UseAtomicUnit)? 0.529177 : 1.0;

    return ($k*$Ra, $k*$Rb, $k*$Rc, $Ralpha, $Rbeta, $Rgamma);
}

sub CalculateHKLInterplanarSpacing
{
    my ($this, $h, $k, $l) = @_;
    Rg11 = $this->{"Rg11"};
    Rg12 = $this->{"Rg12"};
    Rg13 = $this->{"Rg13"};
    Rg21 = $this->{"Rg21"};
    Rg22 = $this->{"Rg22"};
    Rg23 = $this->{"Rg23"};
    Rg31 = $this->{"Rg31"};
    Rg32 = $this->{"Rg32"};
    Rg33 = $this->{"Rg33"};
#print "($h,$k,$l)-($Rg11,$Rg22,$Rg33,$Rg12,$Rg23,$Rg13)<br>\n";
    d = $Rg11*$h*$h + $Rg22*$k*$k + $Rg33*$l*$l
          + 2.0*$Rg12*$h*$k + 2.0*$Rg23*$k*$l + 2.0*$Rg13*$l*$h;
#print "d=$d<br>";
#print "enter: $d ($h,$k,$l),($Rg11,$Rg22,$Rg33,$Rg12,$Rg23,$Rg13)<br>";
    return 0.0 if($d == 0.0);
    $d = sqrt(1.0 / $d);
#print "d=$d<br>";
    return $d;
}

sub CalculateDiffractionAngleFromInterplanarSpacing
{
    my ($this, $wl, $d) = @_;
    return 0.0 if($d == 0.0);
    sq = $wl * 0.5 / $d;
#print "sq: $sq<br>";
    return 0.0 if(abs($sq) > 1.0);
    Q = Sci::dasin($sq);
    return $Q;
}


sub LatticeVectorsByMatrix
{
    my ($this, $UseAtomicUnit) = @_;
    my @v = $this->LatticeVectors($UseAtomicUnit);
    my @a1 = ($v[0], $v[1], $v[2]);
    my @a2 = ($v[3], $v[4], $v[5]);
    my @a3 = ($v[6], $v[7], $v[8]);
    return (\@a1, \@a2, \@a3);
}

# @v in Math::Vector::Real
sub LatticeVectorsByVectors
{
    my ($this, $UseAtomicUnit) = @_;
    my @v = $this->LatticeVectors($UseAtomicUnit);
    a1 = V($v[0], $v[1], $v[2]);
    a2 = V($v[3], $v[4], $v[5]);
    a3 = V($v[6], $v[7], $v[8]);
    return ($a1, $a2, $a3);
}


sub ReciprocalLatticeVectorsByMatrix
{
    my ($this, $UseAtomicUnit) = @_;
    my @v = $this->ReciprocalLatticeVectors($UseAtomicUnit);
    my @a1 = $v[0,2];
    my @a2 = $v[3,5];
    my @a3 = $v[6,8];
    return (\@a1, \@a2, \@a3);
}

sub ReciprocalLatticeVectorsByVectors
{
    my ($this, $UseAtomicUnit) = @_;
    my @v = $this->ReciprocalLatticeVectors($UseAtomicUnit);
    a1 = V($v[0,2]);
    a2 = V($v[3,5]);
    a3 = V($v[6,8]);
    return ($a1, $a2, $a3);
}

sub ReciprocalLatticeVectors
{
    my ($this, $UseAtomicUnit) = @_;
    k = ($UseAtomicUnit)? 1.0 / 0.529177 : 1.0;

    return ($k*$this->{Ra11}, $k*$this->{Ra12}, $k*$this->{Ra13},
            k*$this->{Ra21}, $k*$this->{Ra22}, $k*$this->{Ra23},
            k*$this->{Ra31}, $k*$this->{Ra32}, $k*$this->{Ra33});
}

sub GetYRangeForIteration
{
    my ($this, $ix) = @_;

    ls = lc $this->LatticeSystem();
#print "ls: $ls\n";
    if($ls eq 'cubic' or $ls =~ /(rhomb|trig)/) {
        return $ix;
    }
    elsif($ls =~ /(rhomb|trig)/) {
        return $ix;
    }
    elsif($ls =~ /tetra/ or $ls =~ /hex/) {
        return $ix;
    }
    elsif($ls =~ /ortho/) {
        return 0;
    }
    elsif($ls =~ /mono/) {
        return 0;
    }
    else {
        return 0;
    }
}

sub GetZRangeForIteration
{
    my ($this, $ix, $iy) = @_;

    ls = lc $this->LatticeSystem();
#print "ls: $ls\n";
    if($ls eq 'cubic' or $ls =~ /(rhomb|trig)/) {
        return $iy;
    }
    elsif($ls =~ /tetra/ or $ls =~ /hex/) {
        return 0;
    }
    elsif($ls =~ /ortho/) {
        return 0;
    }
    elsif($ls =~ /mono/) {
        return 0;
    }
    else {
        return 0;
    }
}

sub ConvertLatticeIndexByLatticeSystem
{
    my ($this, $ix, $iy, $iz, $SortDirection) = @_;
    $SortDirection = 1 if(!defined $SortDirection);

    if($SortDirection == 0) {
        return (0, $ix, $iy, $iz);
    }
    
    ls = lc $this->LatticeSystem();
#print "ls: $ls\n";
    IsConverted = 0;
    if($ls eq 'cubic' or $ls =~ /(rhomb|trig)/) {
        IsConverted = 1 if($ix > $iy or $iy > $iz or $ix > $iz);
        if($SortDirection > 0) {
            return ($IsConverted, sort { return $a <=> $b; } ($ix, $iy, $iz) );
        }
        return ($IsConverted, sort { return $b <=> $a; } ($ix, $iy, $iz) );
    }
    elsif($ls =~ /tetra/ or $ls =~ /hex/) {
        IsConverted = 1 if($ix > $iy);
        my @a;
        if($SortDirection > 0) {
            @a =  sort { return $a <=> $b; } ($ix, $iy);
        }
        else {
            @a =  sort { return $b <=> $a; } ($ix, $iy);
        }
        return ($IsConverted, @a, $iz);
    }
    else {
#        die "\nCrystaObject::ConvertLatticeIndexByLatticeSystem: Error: Lattice system [$ls] is not supported.\n";
#        print "\nCrystaObject::ConvertLatticeIndexByLatticeSystem: Warning: Lattice system [$ls] is not supported.\n";
    }
    return ($IsConverted, $ix, $iy, $iz);
}


sub GetPerpendicularHKL
{
    my ($this, $h1, $k1, $l1, $h2, $k2, $l2) = @_;

    my @v1 = $this->GetVectorFromHKL($h1, $k1, $l1);
    my @v2 = $this->GetVectorFromHKL($h2, $k2, $l2);
    my @v3 = Sci::VectorProduct(\@v1, \@v2);
#    my @v3 = ($v1[1]*$v2[2] - $v1[2]*$v2[1],
#              $v1[2]*$v2[0] - $v1[0]*$v2[2],
#              $v1[0]*$v2[1] - $v1[1]*$v2[0]);
#printf "v1=(%7.5f, %7.5f, %7.5f)\n", @v1;
#printf "v2=(%7.5f, %7.5f, %7.5f)\n", @v2;
#printf "v3=(%7.5f, %7.5f, %7.5f)\n", @v3;

    R = new Math::MatrixReal(3, 3);
    $R->assign(1, 1, $this->{Ra11});
    $R->assign(1, 2, $this->{Ra21});
    $R->assign(1, 3, $this->{Ra31});
    $R->assign(2, 1, $this->{Ra12});
    $R->assign(2, 2, $this->{Ra22});
    $R->assign(2, 3, $this->{Ra32});
    $R->assign(3, 1, $this->{Ra13});
    $R->assign(3, 2, $this->{Ra23});
    $R->assign(3, 3, $this->{Ra33});
#print "R=$R\n";
    RR = $R->inverse();
#print "RR=$RR\n";
    my @hkl = ($RR->element(1,1)*$v3[0] + $RR->element(1,2)*$v3[1] + $RR->element(1,3)*$v3[2],
               $RR->element(2,1)*$v3[0] + $RR->element(2,2)*$v3[1] + $RR->element(2,3)*$v3[2],
               $RR->element(3,1)*$v3[0] + $RR->element(3,2)*$v3[1] + $RR->element(3,3)*$v3[2]);
    my ($min, $max) = Utils::CalMinMax(\@hkl, 1);
#printf "hkl=%6.3f %6.3f %6.3f\n", @hkl;
    $min = 1.0e30;
    for(my $i = 0 ; $i < 3 ; $i++) {
        hkl[$i] = 0.0 if(abs($hkl[$i]) < $max * 1.0e-15);
        min = $hkl[$i] if($hkl[$i] != 0 and $min > $hkl[$i]);
    }
    if($min < 1.0e30) {
        for(my $i = 0 ; $i < 3 ; $i++) {
            hkl[$i] /= $min;
        }
    }
#printf "hkl=%6.3f %6.3f %6.3f\n", @hkl;
#my @v32 =  $this->GetVectorFromHKL(@hkl);
#printf "v3=(%7.5f, %7.5f, %7.5f)\n", @v32;
    return @hkl;
}

sub GetVectorFromHKL
{
    my ($this, $h, $k, $l) = @_;
    my @V = ($this->{Ra11} * $h + $this->{Ra21} * $k + $this->{Ra31} * $l,
             $this->{Ra12} * $h + $this->{Ra22} * $k + $this->{Ra32} * $l,
             $this->{Ra13} * $h + $this->{Ra23} * $k + $this->{Ra33} * $l);
#print "hkl=$h,$k,$l: ", join(',', @V), ")\n";
    return @V;
}

sub RMetrics
{
    my ($this) = @_;
    Rg11 = $this->{Rg11};
    Rg12 = $this->{Rg12};
    Rg13 = $this->{Rg13};
    Rg21 = $this->{Rg21};
    Rg22 = $this->{Rg22};
    Rg23 = $this->{Rg23};
    Rg31 = $this->{Rg31};
    Rg32 = $this->{Rg32};
    Rg33 = $this->{Rg33};
    return ($Rg11, $Rg12, $Rg13,
        Rg21, $Rg22, $Rg23,
        Rg31, $Rg32, $Rg33)
}

# @v in Math::Vector::Real
sub SetLatticeVectorsByVectors
{
    my ($this, @v) = @_;
#print "v1=", $v[0], "\n";
#print "v2=", $v[1], "\n";
#print "v3=", $v[2], "\n";
    $this->SetLatticeVector(
        v[0][0], $v[0][1], $v[0][2],
        v[1][0], $v[1][1], $v[1][2],
        v[2][0], $v[2][1], $v[2][2]);
}

sub SetLatticeVectors
{
    my ($this, $a11, $a12, $a13,
        $a21, $a22, $a23, $a31, $a32, $a33) = @_;
    return $this->SetLatticeVector(
        a11, $a12, $a13,
        a21, $a22, $a23, 
        a31, $a32, $a33);
}

sub SetLatticeVector
{
    my ($this, $a11, $a12, $a13,
        $a21, $a22, $a23, $a31, $a32, $a33) = @_;
#print "($this, $a11, $a12, $a13, $a21, $a22, $a23, $a31, $a32, $a33)\n";
    $this->{a11} = $a11;
    $this->{a12} = $a12;
    $this->{a13} = $a13;
    $this->{a21} = $a21;
    $this->{a22} = $a22;
    $this->{a23} = $a23;
    $this->{a31} = $a31;
    $this->{a32} = $a32;
    $this->{a33} = $a33;
#print "SetLatticeVector: a1=$this->{a11},$this->{a12},$this->{a13}\n";
#print "SetLatticeVector: a2=$this->{a21},$this->{a22},$this->{a23}\n";
#print "SetLatticeVector: a3=$this->{a31},$this->{a32},$this->{a33}\n";
    $this->CalMetricsFromLatticeVector();
    $this->CalLatticeParametersFromMetrics();
    return;
}

sub SetMetrics
{
    my ($this,$g11,$g12,$g13,$g21,$g22,$g23,$g31,$g32,$g33)
        = @_;
    $this->{'g11'} = $g11;
    $this->{'g12'} = $g12;
    $this->{'g13'} = $g13;
    $this->{'g21'} = $g21;
    $this->{'g22'} = $g22;
    $this->{'g23'} = $g23;
    $this->{'g31'} = $g31;
    $this->{'g32'} = $g32;
    $this->{'g33'} = $g33;
    return ($this,$g11,$g12,$g13,$g21,$g22,$g23,$g31,$g32,$g33);
}

sub CalLatticeParametersFromMetrics
{
    my ($this) = @_;

    a = sqrt($this->{'g11'});
    b = sqrt($this->{'g22'});
    c = sqrt($this->{'g33'});
    alpha = $this->{'g23'} / $b / $c;
       $alpha = Sci::dacos($alpha);
    beta  = $this->{'g13'} / $a / $c;
       $beta  = Sci::dacos($beta);
    gamma = $this->{'g12'} / $a / $b;
       $gamma = Sci::dacos($gamma);
    $a = Sci::Round($a, 6);
    $b = Sci::Round($b, 6);
    $c = Sci::Round($c, 6);
#print "a,b,c=$a,$b,$c\n";
    $alpha = Sci::Round($alpha, 4);
    $beta  = Sci::Round($beta , 4);
    $gamma = Sci::Round($gamma, 4);
#print "alpha=$alpha,$beta,$gamma\n";
    $this->SetLatticeParameters($a,$b,$c,$alpha,$beta,$gamma);
}

sub CalMetricsFromLatticeVector
{
    my ($this) = @_;
    my ($a11, $a12, $a13, $a21, $a22, $a23, $a31, $a32, $a33) = (
        this->{a11}, $this->{a12}, $this->{a13},
        this->{a21}, $this->{a22}, $this->{a23},
        this->{a31}, $this->{a32}, $this->{a33});

    $this->{g11} = $a11*$a11 + $a12*$a12 + $a13*$a13;
    $this->{g22} = $a21*$a21 + $a22*$a22 + $a23*$a23;
    $this->{g33} = $a31*$a31 + $a32*$a32 + $a33*$a33;
    $this->{g12} = $this->{g21} = $a11*$a21 + $a12*$a22 + $a13*$a23;
    $this->{g23} = $this->{g32} = $a21*$a31 + $a22*$a32 + $a23*$a33;
    $this->{g31} = $this->{g13} = $a31*$a11 + $a32*$a12 + $a33*$a13;
    return ($this->{g11}, $this->{g12}, $this->{g13},
            this->{g21}, $this->{g22}, $this->{g23},
            this->{g31}, $this->{g32}, $this->{g33})
}

sub CalculateLatticeConstantFromVector
{
    my ($this) = @_;
    a11 = $this->{a11};
    a12 = $this->{a12};
    a13 = $this->{a13};
    a21 = $this->{a21};
    a22 = $this->{a22};
    a23 = $this->{a23};
    a31 = $this->{a31};
    a32 = $this->{a32};
    a33 = $this->{a33};
#print "a1 ($a11, $a12, $a13)\n";
#print "a2 ($a21, $a22, $a23)\n";
#print "a3 ($a31, $a32, $a33)\n";
    a = sqrt($a11*$a11 + $a12*$a12 + $a13*$a13);
    b = sqrt($a21*$a21 + $a22*$a22 + $a23*$a23);
    c = sqrt($a31*$a31 + $a32*$a32 + $a33*$a33);
#print "a=$a, $b, $c\n";
    ab = $a11*$a21 + $a12*$a22 + $a13*$a23;
    bc = $a21*$a31 + $a22*$a32 + $a23*$a33;
    ca = $a31*$a11 + $a32*$a12 + $a33*$a13;
#print "ab=$ab, $bc, $ca\n";
#exit;
    cosg = $ab / $a / $b;
    cosa = $bc / $b / $c;
    cosb = $ca / $c / $a;
    alpha = Sci::dacos($cosa);
    beta  = Sci::dacos($cosb);
    gamma = Sci::dacos($cosg);
    $a = Utils::Round($a+0.0, 6); #sprintf("%11.6f", $a) + 0.0;
    $b = Utils::Round($b+0.0, 6); #sprintf("%11.6f", $b) + 0.0;
    $c = Utils::Round($c+0.0, 6); #sprintf("%11.6f", $c) + 0.0;
    $alpha = Utils::Round($alpha+0.0, 6); #sprintf("%12.6f", $alpha) + 0.0;
    $beta  = Utils::Round($beta+0.0, 6); #sprintf("%12.6f", $beta) + 0.0;
    $gamma = Utils::Round($gamma+0.0, 6); #sprintf("%12.6f", $gamma) + 0.0;
#print "a=$a $b $c $alpha $beta $gamma\n";
    return ($a,$b,$c,$alpha,$beta,$gamma);
}

sub FindAtomTypeByName
{
    my ($this, $atomname) = @_;
    atomlistref = $this->{"RefAtomTypeList"};

    IsPartial = 0;
    for(my $i = 0 ; $i < @$atomlistref ; $i++) {
        type = $atomlistref->[$i];
        name = $type->Label();
#print "i=$i: name=$name\n";
        if($name =~ /\[.*?\]/) {
            IsPartial = 1;
            last;
        }
    }
#print "IsP=$IsPartial\n";

    if(!$IsPartial) {
        atomname =~ s/([\.\d+-]+)//;
    }
    for(my $i = 0 ; $i < @$atomlistref ; $i++) {
        type = $atomlistref->[$i];
        name;
        if($IsPartial) {
            name = $type->Label();
        }
        else {
            name = $type->AtomNameOnly();
            name =~ s/([\d+-]+)//;
        }
#print "name: $atomname <=> $name\n";
        return $type if(uc $name eq uc $atomname);
    }
    return undef;
}

sub SplitAtomType
{
    my ($this) = @_;

    my @AtomTypeList = $this->GetCAtomTypeList();
    my @AtomSiteList = $this->GetCAsymmetricAtomSiteList();

    $this->ClearAtomTypeList();
    $this->ClearAtomSiteList();
    my %at;
#my $n = @AtomSiteList;
#print "nsite=$n\n";
    for(my $i = 0 ; $i < @AtomSiteList ; $i++) {
        site   = $AtomSiteList[$i];
        name   = $site->AtomNameOnly();
        label  = $site->Label();
        charge = $site->Charge();
        occ    = $site->Occupancy();
        my ($x, $y, $z)  = $site->Position();
        my ($fx,$fy,$fz) = $site->Force();

        at{$name}++;
        name2 = "$name\{" . Utils::IntToAlpha($at{$name} - 1) . "\}$charge";
#print "name2=$name2\n";
        this->AddAtomSite($label, $name2, $x, $y, $z, $occ, $fx, $fy, $fz);
#        this->AddAtomType($name2, 1);
#print "Site (indep type) $i: $name2 (Z=$charge) (occ=$occ): $key\n";
    }
    $this->ExpandCoordinates();
}

sub SplitPartialSites
{
    my ($this) = @_;

    my @AtomTypeList = $this->GetCAtomTypeList();
    my @AtomSiteList = $this->GetCAsymmetricAtomSiteList();
#    my @ExpandedAtomSiteList = $this->GetCExpandedAtomSiteList();

    $this->ClearAtomTypeList();
    $this->ClearAtomSiteList();
    my %at;
    for(my $i = 0 ; $i < @AtomSiteList ; $i++) {
        site   = $AtomSiteList[$i];
        name   = $site->AtomNameOnly();
        label  = $site->Label();
        charge = $site->Charge();
        occ    = $site->Occupancy();
        my ($x, $y, $z)  = $site->Position();
        my ($fx,$fy,$fz) = $site->Force();
        key;
        if(abs($occ - 1.0) < 1.0e-4) {
            key = "$name$charge";
        }
        else {
            key = sprintf("$name${charge}[%6.4f]", $occ);
        }
        this->AddAtomSite($label, $key, $x, $y, $z, $occ, $fx, $fy, $fz);

        next if($at{$key});
        at{$key}++;
        this->AddAtomType($key, 1);
print "Site (indep type) $i: $name (Z=$charge) (occ=$occ): $key\n";
    }

    my @ATL = $this->GetCAtomTypeList();
    for(my $i = 0 ; $i < @ATL ; $i++) {
        type    = $ATL[$i];
        nameraw = $type->Label();
        name    = $type->AtomNameOnly();
        charge  = $type->Charge();
        occ     = $type->Occupancy();
        key     = sprintf("$name${charge}[%6.4f]", $occ);

print "Type $i: {$nameraw} $name (Z=$charge) (occ=$occ): $key\n";
    }

    my @ASL = $this->GetCAsymmetricAtomSiteList();
    for(my $i = 0 ; $i < @ASL ; $i++) {
        site    = $ASL[$i];
        nameraw = $site->AtomName();
        name    = $site->AtomNameOnly();
        charge  = $site->Charge();
        occ     = $site->Occupancy();
        key     = sprintf("$name${charge}[%6.4f]", $occ);
print "c $i: {$nameraw}: $name (Z=$charge) (occ=$occ): $key\n";
    }

    $this->ExpandCoordinates();
#exit;
}

sub TotalCharge
{
    my ($this) = @_;
    
    $this->ExpandCoordinates();
    my @ExpandedAtomSiteList   = $this->GetCExpandedAtomSiteList();
    nTotalExpandedAtomSite = $this->nTotalExpandedAtomSite();

    Z = 0.0;
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        atom      = $ExpandedAtomSiteList[$i];
        charge    = $atom->Charge();
        occ       = $atom->Occupancy();
        Z += $occ * $charge;
    }
    return $Z;
}

sub CheckTotalCharge
{
    my ($this, $EPS) = @_;
    $EPS = 1.0e-3 if($EPS eq '');
    Z = $this->TotalCharge();
    if(abs($Z) > $EPS) {
        print "Error in CrystalObject::CheckTotalCharge: Total charge ($Z) is greater than EPS=$EPS.\n";
        exit 0;
    }
    return 1;
}

sub GetCBravaisLatticeAsymmetricAtomSiteList
{
    my ($this) = @_;
    $this->ExpandCoordinates();
    SPG = $this->GetCSpaceGroup();
    my @ExpandedAtomSiteList   = $this->GetCExpandedAtomSiteList();
    nTotalExpandedAtomSite = $this->nTotalExpandedAtomSite();

    my (@Tx, @Ty, @Tz);
    nT = $SPG->nTranslation();
    for(my $it = 1 ; $it <= $nT ; $it++) {
        ($Tx[$it-1], $Ty[$it-1], $Tz[$it-1]) = $SPG->TranslationVector($it);
    }

    my @AtomSite;
    c = 0;
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        atom      = $ExpandedAtomSiteList[$i];
        label     = $atom->Label();
        atomname  = $atom->AtomNameOnly();
        charge    = $atom->Charge();
        my ($x,$y,$z) = $atom->Position();
        occ       = $atom->Occupancy();
#print "$i: $atomname $charge [$label]: ($x, $y, $z) [occ=$occ]\n";

        IsExist = 0;
        for(my $it = 1 ; $it < $nT ; $it++) {
            my ($tx, $ty, $tz) = ($Tx[$it], $Ty[$it], $Tz[$it]);
            x1 = Utils::Reduce01($x + $tx);
            y1 = Utils::Reduce01($y + $ty);
            z1 = Utils::Reduce01($z + $tz);
            for(my $j = 0 ; $j < $c ; $j++) {
                atom2 = $AtomSite[$j];
                my ($x2, $y2, $z2) = $atom2->Position();
                dis = $this->GetNearestInterAtomicDistance($x1, $y1, $z1, $x2, $y2, $z2);
#print "$i-$it-$j: ($x,$y,$z): ($x1,$y1,$z1)-($x2,$y2,$z2)=$dis\n";
                if($dis < 1.0e-6) {
                    IsExist = 1;
                    last;
                }
            }
            last if($IsExist);
        }
        if(!$IsExist) {
            AtomSite[$c] = $atom;
            c++;
        }
    }

    return @AtomSite;
}

sub GetimaxByRmax
{
    my ($this, $Rmax) = @_;

    my ($a,$b,$c,$alpha,$beta,$gamma) = $this->LatticeParameters();
    return (int($Rmax/$a)+1, int($Rmax/$b)+1, int($Rmax/$c)+1);
}

sub FindCoordinationsForAllAtoms
{
# $iTargetSites: 0: Expanded  1: Asymmetric sites
    my ($Crystal, $iTargetSites, $CoordRMax, $IsPrint) = @_;
    $IsPrint = 0 if(!defined $IsPrint);

    my (@SiteList, $nSite);
    if($iTargetSites == 0) {
        @SiteList = $Crystal->GetCExpandedAtomSiteList();
    }
    else {
        @SiteList = $Crystal->GetCAsymmetricAtomSiteList();
    }
    $nSite    = @SiteList;

    for(my $i = 0 ; $i < $nSite ; $i++) {
        pAtom0       = $SiteList[$i];
        label0       = $pAtom0->Label();
        type0        = $pAtom0->AtomName();
        my ($x0,$y0,$z0) = $pAtom0->Position(1);
#        occupancy0   = $pAtom0->Occupancy();
#        iAtomType0   = $Crystal->FindiAtomType($type);
        id0          = ($iTargetSites == 0)? $i : $pAtom0->{IdAsymmetricAtomSite} - 1;

print "Coordination arround $i-th atom [$label0: $type0 (id=$id0)] ($x0, $y0, $z0)\n" if($IsPrint);
        pAtom0->{idSite}= ($iTargetSites == 0)? $i : $pAtom0->{IdAsymmetricAtomSite} - 1;
        Crystal->FindCoordinationsAroundAtom($pAtom0, $iTargetSites, $CoordRMax);

        idSite0 = $pAtom0->{idSite};
        pCoord  = $pAtom0->{pCoordAtoms};
        for(my $i = 0 ; $i < @$pCoord ; $i++) {
            pCA            = $pCoord->[$i];
            iSite1         = $pCA->{iSite};
            idSite1        = $pCA->{idSite};
            label1         = $pCA->Label();
            type1          = $pCA->AtomName();
            my ($x1, $y1, $z1) = $pCA->Position();
            my ($X1, $Y1, $Z1) = @{$pCA->{pCoord}};
            r              = $pCA->{r};
printf "  %03d: [%3s: %4s (id = %03d)] (%8.4f, %8.4f, %8.4f): r = %8.4f  frac:(%6.3f, %6.3f, %6.3f)\n",
        iSite1, $label1, $type1, $idSite1, $X1, $Y1, $Z1, $r, $x1, $y1, $z1 if($IsPrint);
        }
    }
    
    return \@SiteList;
}

sub FindCoordinationsAroundAtom
{
    my ($Crystal, $pAtom0, $iTargetSites, $CoordRMax) = @_;

    my ($x0,$y0,$z0) = $pAtom0->Position(1);
    my ($ixmax, $iymax, $izmax) = $Crystal->GetimaxByRmax($CoordRMax);

    my @ExpandedAtomSiteList   = $Crystal->GetCExpandedAtomSiteList();
    nTotalExpandedAtomSite = $Crystal->nTotalExpandedAtomSite();

    my @CoordAtoms;
    c = 0;
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        pAtom1       = $ExpandedAtomSiteList[$i];
        label1       = $pAtom1->Label();
        type1        = $pAtom1->AtomName();
        id1          = ($iTargetSites == 0)? $i : $pAtom1->{IdAsymmetricAtomSite} - 1;
        my ($x1,$y1,$z1) = $pAtom1->Position(1);

        my @dislist = $Crystal->GetNearestInterAtomicDistances(
                        x0, $y0, $z0, $x1, $y1, $z1, $CoordRMax, 0, $ixmax, $iymax, $izmax);

my $p  = $ExpandedAtomSiteList[$id1];
my $n1 = $pAtom1->AtomName();
my $n  = $p->AtomName();
my $l  = $p->Label();
if($n1 ne $n) {
    print "Error: Atom name mismatch ($n1 ne $n) for ($label1 vs $l) (i=$i vs id=$id1)\n";
    exit;
}

        for(my $j = 0 ; $j < @dislist ; $j++) {
            piCell = $dislist[$j]->{piCell};
            r      = $dislist[$j]->{r};

            CoordAtoms[$c] = new AtomSite;
            %{$CoordAtoms[$c]} = %$pAtom1;
            CoordAtoms[$c]->{pSite}  = $pAtom1;
            CoordAtoms[$c]->{iSite}  = $i;
            CoordAtoms[$c]->{idSite} = $id1;
            CoordAtoms[$c]->{piCell} = $piCell;
            CoordAtoms[$c]->{pCoord} = [$x1 + $piCell->[0], $y1 + $piCell->[1], $z1 + $piCell->[2]];
            CoordAtoms[$c]->{r}      = $r;

            l = $CoordAtoms[$c];
#print "  $i: [$label: $type (id = $id1)] ($l->{pCoord}[0], $l->{pCoord}[1], $l->{pCoord}[2]): r = $l->{r}\n";
            c++;
        }
    }

    $pAtom0->{pCoordAtoms} = \@CoordAtoms;

    return $pAtom0;
}

sub GetNearestInterAtomicDistances
{
    my ($this, $x0, $y0, $z0, $x1, $y1, $z1, $CoordRMax, $AllowZero, $ixmax, $iymax, $izmax) = @_;
    $AllowZero = 1 if(!defined $AllowZero);
    if(!defined $ixmax) {
        ($ixmax, $iymax, $izmax) = $this->GetimaxByRmax($CoordRMax);
    }
    my @list;
    c = 0;
    for(my $ix = -$ixmax ; $ix <= $ixmax ; $ix++) {
    for(my $iy = -$iymax ; $iy <= $iymax ; $iy++) {
    for(my $iz = -$izmax ; $iz <= $izmax ; $iz++) {
        r = $this->GetInterAtomicDistance($x0, $y0, $z0, $x1+$ix, $y1+$iy, $z1+$iz);
#print "$ix,$iy,$iz: r=$r\n";
        if($r < $CoordRMax) {
            next if(!$AllowZero and $r < 1.0e-5);

            list[$c++] = {
                piCell => [$ix, $iy, $iz],
                r      => $r,
                };
        }
    }
    }
    }
    return @list;
}

sub GetNearestInterAtomicDistance
{
    my ($this, $x0, $y0, $z0, $x1, $y1, $z1, $AllowZero) = @_;
    $AllowZero = 1 if(!defined $AllowZero);

    $x0 = $this->FindNearestEquivalentFractionCoordinate($x0, $x1);
    $y0 = $this->FindNearestEquivalentFractionCoordinate($y0, $y1);
    $z0 = $this->FindNearestEquivalentFractionCoordinate($z0, $z1);
    r = $this->GetInterAtomicDistance($x0, $y0, $z0, $x1, $y1, $z1);
#print "r=$r\n";
    if(!$AllowZero and $r <= 1.0e-5) {
        r = 1.0e10;
        for(my $ix = -1 ; $ix <= 1 ; $ix++) {
        for(my $iy = -1 ; $iy <= 1 ; $iy++) {
        for(my $iz = -1 ; $iz <= 1 ; $iz++) {
            r2 = $this->GetInterAtomicDistance($x0, $y0, $z0, $x1+$ix, $y1+$iy, $z1+$iz);
#print "$ix,$iy,$iz: r2=$r2\n";
            r = $r2 if($r2 > 1.0e-5 and $r > $r2);
        }
        }
        }
    }
    
    return $r;
}

sub GetInterAtomicDistance
{
    my ($this, $x0, $y0, $z0, $x1, $y1, $z1) = @_;;
    dx = $x1 - $x0;
    dy = $y1 - $y0;
    dz = $z1 - $z0;
    g11 = $this->{"g11"};
    g22 = $this->{"g22"};
    g33 = $this->{"g33"};
    g12 = 2.0* $this->{"g12"};
    g13 = 2.0* $this->{"g13"};
    g23 = 2.0* $this->{"g23"};
    r2 = $g11*$dx*$dx + $g22*$dy*$dy + $g33*$dz*$dz
           + $g12*$dx*$dy + $g13*$dx*$dz + $g23*$dy*$dz;
    return sqrt($r2);
}

sub GetInterAtomicAngle
{
    my ($this, $x0, $y0, $z0, $x1, $y1, $z1,
           $x2, $y2, $z2) = @_;;

    dis01 = $this->GetInterAtomicDistance($x0,$y0,$z0,$x1,$y1,$z1);
    return 0.0 if($dis01 == 0.0);
    dis02 = $this->GetInterAtomicDistance($x0,$y0,$z0,$x2,$y2,$z2);
    return 0.0 if($dis02 == 0.0);
#print "dis: $dis01, $dis02\n";
    dx01 = $x1 - $x0;
    dy01 = $y1 - $y0;
    dz01 = $z1 - $z0;
    dx02 = $x2 - $x0;
    dy02 = $y2 - $y0;
    dz02 = $z2 - $z0;
    g11 = $this->{"g11"};
    g22 = $this->{"g22"};
    g33 = $this->{"g33"};
    g12 = 2.0* $this->{"g12"};
    g13 = 2.0* $this->{"g13"};
    g23 = 2.0* $this->{"g23"};
    ip  = $g11*$dx01*$dx02 + $g22*$dy01*$dy02 + $g33*$dz01*$dz02
        + $g12*$dx01*$dy02 + $g13*$dx01*$dz02
        + $g12*$dy01*$dx02 + $g23*$dy01*$dz02
        + $g13*$dz01*$dx02 + $g23*$dz01*$dy02;

    cosa = $ip / $dis01 / $dis02;
#print "cos: $cosa<br>";
    angle = Sci::dacos($cosa);
#print "cos: $cosa $angle<br>";
    $angle = 360.0 - $angle if($angle > 180.0001);

    return $angle;
}

#iAtomTypeは1から始まる
sub FindiAtomType
{
    my ($this, $atomname) = @_;
#print "F:$atomname\n";
    $atomname =~ s/^(\D+)(\d.*)$/$1/;
    $atomname = uc $atomname;
#print "F:$atomname\n";
    my @AtomTypeList = $this->GetCAtomTypeList();
    nAtomType = @AtomTypeList;
    for(my $i = 0 ; $i < $nAtomType ; $i++) {
        name = $AtomTypeList[$i]->AtomName();
        name =~ s/^(\D+)(\d.*)$/$1/;
#print "$i: $name\n";
        return $i+1 if($atomname eq uc $name);
    }
    return -1;
}

sub CalCartesianCoordinate
{
    my ($this, $x, $y, $z) = @_;

    my ($ax, $ay, $az, $bx, $by, $bz, $cx, $cy, $cz) = $this->LatticeVectorsByOutputMode(0);
    rx = $ax * $x + $bx * $y + $cx * $z;
    ry = $ay * $x + $by * $y + $cy * $z;
    rz = $az * $x + $bz * $y + $cz * $z;
    return ($rx, $ry, $rz);
}

sub CalFractionalCoordinate
{
    my ($this, $rx, $ry, $rz) = @_;

    my ($ax, $ay, $az, $bx, $by, $bz, $cx, $cy, $cz) = $this->LatticeVectorsByOutputMode(0);
    T = new Math::MatrixReal(3, 3);
    $T->assign(1, 1, $ax);
    $T->assign(1, 2, $bx);
    $T->assign(1, 3, $cx);
    $T->assign(2, 1, $ay);
    $T->assign(2, 2, $by);
    $T->assign(2, 3, $cy);
    $T->assign(3, 1, $az);
    $T->assign(3, 2, $bz);
    $T->assign(3, 3, $cz);
    RT = $T->inverse();

    x = $RT->element(1,1) * $rx + $RT->element(1,2) * $ry + $RT->element(1,3) * $rz;
    y = $RT->element(2,1) * $rx + $RT->element(2,2) * $ry + $RT->element(2,3) * $rz;
    z = $RT->element(3,1) * $rx + $RT->element(3,2) * $ry + $RT->element(3,3) * $rz;
    return ($x, $y, $z);
}

sub FindIdenticalSite
{
    my ($this, $atomname, $x, $y, $z, $TolD, $iSiteStart) = @_;
    $TolD = 0.01 if(!defined $TolD);
    $iSiteStart = 0 if(!defined $iSiteStart);

    $atomname =~ s/\d[\+\-]?//;
    my @AtomSiteList = $this->GetCExpandedAtomSiteList();
    nSite        = @AtomSiteList;
    for(my $ia = $iSiteStart ; $ia < $nSite ; $ia++) {
        site = $AtomSiteList[$ia];
        name = $site->AtomNameOnly();
        next if(uc $name ne uc $atomname);
        my ($x1, $y1, $z1) = $site->Position();

        dis = $this->GetNearestInterAtomicDistance($x, $y, $z, $x1, $y1, $z1);
        return $ia if($dis < $TolD);
    }
    return undef;
}

sub FindIdenticalAsymmetricSite
{
    my ($this, $atomname, $x, $y, $z) = @_;
    $atomname =~ s/\d[\+\-]?//;
    my @AtomSiteList = $this->GetCAsymmetricAtomSiteList();
    nAsymmetricAtomSite = @AtomSiteList;
    for(my $ia = 0 ; $ia < $nAsymmetricAtomSite ; $ia++) {
        site = $AtomSiteList[$ia];
        name = $site->AtomNameOnly();
        next if(uc $name ne uc $atomname);
        my ($x1, $y1, $z1) = $site->Position();

        dis = $this->GetNearestInterAtomicDistance($x, $y, $z, $x1, $y1, $z1);
        return $ia if($dis < 0.01);
    }
    return undef;
}

sub FindMass
{
    my ($this, $atomname) = @_;

    my @AtomTypeList = $this->GetCAtomTypeList();
    for(my $i = 0 ; $i < @AtomTypeList ; $i++) {
        atom = $AtomTypeList[$i];
        name = $atom->AtomNameOnly();
        next if(uc $name ne uc $atomname);
        return $atom->AtomicMass();
    }
    return undef;
}

sub ConvertLattice
{
    my ($this, $T, $LatticeParameterOnly, $CheckConsistency, $IsPrint) = @_;
    nTotalExpandedAtomSite = $this->nTotalExpandedAtomSite();

    SPG     = $this->GetCSpaceGroup();
    SPGName = $SPG->SPGName();
    my ($a, $b, $c, $alpha, $beta, $gamma) = $this->LatticeParameters();

#実空間座標(x,y,z)の変換行列
    tRT = new Math::MatrixReal(3, 3);
    $tRT->transpose($T);
    $tRT = $tRT->inverse();

    g = new Math::MatrixReal(3, 3);
    $g->assign(1, 1, $this->{'g11'});
    $g->assign(1, 2, $this->{'g12'});
    $g->assign(1, 3, $this->{'g13'});
    $g->assign(2, 1, $this->{'g21'});
    $g->assign(2, 2, $this->{'g22'});
    $g->assign(2, 3, $this->{'g23'});
    $g->assign(3, 1, $this->{'g31'});
    $g->assign(3, 2, $this->{'g32'});
    $g->assign(3, 3, $this->{'g33'});

    g2 = new Math::MatrixReal(3, 3);
    if(0) {
#    if($SPGName =~ /^\s*[FIABC]/i) {
        for(my $i = 1 ; $i <= 3 ; $i++) {
        for(my $j = 1 ; $j <= 3 ; $j++) {
            tij = $T->element($i,$j);
#print "$i,$j: $tij\n";
            g2->assign($i, $j, $tij);
        }
        }
    }
    else {
        for(my $i1 = 1 ; $i1 <= 3 ; $i1++) {
        for(my $i2 = 1 ; $i2 <= 3 ; $i2++) {
            val = 0.0;
            for(my $j1 = 1 ; $j1 <= 3 ; $j1++) {
            for(my $j2 = 1 ; $j2 <= 3 ; $j2++) {
                val += $g->element($j1,$j2) 
                      * $T->element($i1,$j1)
                      * $T->element($i2,$j2);
            }
            }
#print "val: $i1,$i2: $val\n";
            g2->assign($i1, $i2, $val);
        }
        }
    }
#print "SPGName:$SPGName\n";
    $this->{ConventionalSPGName} = $SPGName;
    $this->{ConventionalLatticeParameters} = [$a, $b, $c, $alpha, $beta, $gamma];
print "***ConvL:$a, $b, $c, $alpha, $beta, $gamma\n";

    NewCrystal = new Crystal();
    $NewCrystal->{ConventionalSPGName} = $SPGName;
    $NewCrystal->{ConventionalLatticeParameters} = [$a, $b, $c, $alpha, $beta, $gamma];
    $NewCrystal->SetSpaceGroup("P 1", 1);
    $NewCrystal->SetMetrics(
        g2->element(1,1), $g2->element(1,2), $g2->element(1,3),
        g2->element(2,1), $g2->element(2,2), $g2->element(2,3),
        g2->element(3,1), $g2->element(3,2), $g2->element(3,3));
    $NewCrystal->CalLatticeParametersFromMetrics();
# Repeat to get proper lattice vectors for Primitive cell reduced by conventional cell
    $NewCrystal->CalcMetrics();
#return $NewCrystal;
#my ($a2, $b2, $c2, $alpha2, $beta2, $gamma2) = $NewCrystal->LatticeParameters();
#die "**a'=$a2 b'=$b2 c'=$c2  alpha'=$alpha2 beta'=$beta2 gamma'=$gamma2\n";

    return $NewCrystal if($LatticeParameterOnly);

    $this->ExpandCoordinates();

    amin = 0;
    amax = 0;
    for(my $i = 1 ; $i <= 3 ; $i++) {
        e = $T->element($i,1);
        amin = int($e-0.999) if($e < $amin);
        amax = int($e-0.001) if($e > $amax);
    }
    bmin = 0;
    bmax = 0;
    for(my $i = 1 ; $i <= 3 ; $i++) {
        e = $T->element($i,2);
        bmin = int($e-0.999) if($e < $bmin);
        bmax = int($e-0.001) if($e > $bmax);
    }
    cmin = 0;
    cmax = 0;
    for(my $i = 1 ; $i <= 3 ; $i++) {
        e = $T->element($i,3);
        cmin = int($e-0.999) if($e < $cmin);
        cmax = int($e-0.001) if($e > $cmax);
    }
    if($CheckConsistency) {
        amax++; $bmax++; $cmax++;
    }
#$bmin = 0;
print "Expansion range: ($amin,$bmin,$cmin) - ($amax,$bmax,$cmax)<br>\n" if($IsPrint);
    my @ExpandedAtomSiteList   = $this->GetCExpandedAtomSiteList();
    $nTotalExpandedAtomSite = $this->nTotalExpandedAtomSite();
    count = 0;
    my %AtomTypeCount;

    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        atom        = $ExpandedAtomSiteList[$i];
        atomnameraw = $atom->AtomName();
        atomname    = $atom->AtomNameOnly();
        charge      = $atom->Charge();
        my ($x,$y,$z)   = $atom->Position(1);
        occ         = $atom->Occupancy();
        key;
        if(abs($occ - 1.0) < 1.0e-4) {
            key = "$atomname$charge";
        }
        else {
            key = sprintf("$atomname${charge}[%6.4f]", $occ);
        }
#print "key-name: $key : $atomnameraw\n";

        for(my $ia = $amin ; $ia <= $amax ; $ia++) {
        for(my $ib = $bmin ; $ib <= $bmax ; $ib++) {
        for(my $ic = $cmin ; $ic <= $cmax ; $ic++) {
            xyz = new Math::MatrixReal(3, 1);
            xyz->assign(1,1, $x+$ia);
            xyz->assign(2,1, $y+$ib);
            xyz->assign(3,1, $z+$ic);
            xyz = $tRT * $xyz;
            x1 = Utils::Reduce01($xyz->element(1,1));
            y1 = Utils::Reduce01($xyz->element(2,1));
            z1 = Utils::Reduce01($xyz->element(3,1));
#print "(",$x+$ia,",",$y+$ib,",",$z+$ic,",) - ($x1,$y1,$z1)<br>";
            WillAdd = 1;
            MinimumDistance = 0.1;
            if($NewCrystal->nAsymmetricAtomSite() == 0) {
                AtomTypeCount{$atomname}++;
                label = $atomname . $AtomTypeCount{$atomname};
                NewCrystal->AddAtomSite($label, $key, $x1, $y1, $z1, $occ);
#                NewCrystal->AddAtomSite($label, $atomname, $x1, $y1, $z1, $occ);
print "Passed0: $label: $key: ($x1,$y1,$z1) occ=$occ<br>\n";
                next;
            }
            my @AtomList = $NewCrystal->GetCAsymmetricAtomSiteList();
            nAtom = @AtomList;
            for(my $j = 0 ; $j < @AtomList ; $j++) {
                atom0         = $AtomList[$j];
#                atom0         = $NewCrystal->GetCAtomSite($j);
                atomnameraw0  = $atom0->AtomName();
                atomname0     = $atom0->AtomNameOnly();
                charge0       = $atom0->Charge();
                my ($x0,$y0,$z0)  = $atom0->Position(1);
                occ0          = $atom0->Occupancy();
#print "is=18 $atomnameraw0 $atomname0\n" if($i == 18);
                key0;
                if(abs($occ0 - 1.0) < 1.0e-4) {
                    key0 = "$atomname0$charge0";
                }
                else {
                    key0 = sprintf("$atomname0${charge0}[%6.4f]", $occ0);
                }
#print "keys(is=$i iatom=$j ($ia,$ib,$ic)): $key $key0\n" if($key =~ /Si/);
                next if($key ne $key0);
#                next if($atomname ne $atomname0);

                dis = $NewCrystal->GetNearestInterAtomicDistance(
                    x0, $y0, $z0, $x1, $y1, $z1);
#print "$nAtom: ($i,$j): ($ia,$ib,$ic): ($x0,$y0,$z0)-($x1,$y1,$z1) ";
#print "dis: $dis<br>";
                if($dis < $MinimumDistance) {
#print "Dropped:     : $key: ($x1,$y1,$z1)-($x0,$y0,$z0) occ=$occ<br>\n";
                    WillAdd = 0;
                    last;
                }
            }
            if($WillAdd) {
                AtomTypeCount{$atomname}++;
                label = $atomname . $AtomTypeCount{$atomname};
#print "Passed: $label: $key: ($x1,$y1,$z1) occ=$occ<br>\n";
                NewCrystal->AddAtomSite($label, $key, $x1, $y1, $z1, $occ);
#                NewCrystal->AddAtomSite($label, $atomname, $x1, $y1, $z1, $occ);
            }
        }
        }
        }
        count++;
    }

    $NewCrystal->ExpandCoordinates();

    return $NewCrystal;
}

sub SetIsSelected
{
    my($this, $index, $IsSelected) = @_;
    RefAtomList = $this->{"RefExpandedAtomSiteList"};
    atom = $RefAtomList->[$index];
    if($atom ne '') {
        return $atom->SetIsSelected($IsSelected);
    }
    return 0;
}

sub IsSelected
{
    my($this, $index) = @_;
    RefAtomList = $this->{"RefExpandedAtomSiteList"};
    atom = $RefAtomList->[$index];
    if($atom ne '') {
        return $atom->IsSelected();
    }
    return 0;
}

sub SameIdSiteIndex
{
    my ($this, $idm1, $sid) = @_;
    SameIdSiteIndex = $this->{'SameIdSiteIndex'};
    return $SameIdSiteIndex->{"[$idm1][$sid]"};
}

sub SortAtomTypeOrder
{
    my ($this, $sorttype) = @_;
#print "<H1>SORT: $sorttype</H1>";
    if($sorttype =~ /ChargeAscend/i) {
        ref = $this->{"RefAtomTypeList"};
        @$ref = sort { $a->Charge() <=> $b->Charge() } @$ref;
    }
    elsif($sorttype =~ /ChargeDescend/i) {
        ref = $this->{"RefAtomTypeList"};
        @$ref = sort { -$a->Charge() <=> -$b->Charge() } @$ref;
    }
    return 1;
}

sub FillAtomTypeData
{
    my ($this) = @_;
    ref = $this->{"RefAtomTypeList"};
    for(my $i = 0 ; $i < @$ref ; $i++) {
        ref->[$i]->FillAtomTypeData();
    }
}

sub ReadASF
{
    my ($this, $Source, $StopByError) = @_;
    $StopByError = 1 if(!defined $StopByError);

    my @AtomTypeList = $this->GetCAtomTypeList();
    nAtomType = @AtomTypeList;
    for(my $i = 0 ; $i < $nAtomType ; $i++) {
        Rietan = new RietanFP;
        AtomName;
        AtomNameOnly = $AtomTypeList[$i]->AtomNameOnly(1);
        Charge       = AtomType::ChargeToVal($AtomTypeList[$i]->Charge());
#print "a1a: $AtomName (Charge=$Charge)\n";
        if($Charge == 1) {
            AtomName = "$AtomNameOnly+";
        }
        elsif($Charge == -1) {
            AtomName = "$AtomNameOnly-";
        }
        elsif($Charge > 1) {
            AtomName = "$AtomNameOnly$Charge+";
        }
        elsif($Charge < -1) {
            Charge = -$Charge;
            AtomName = "$AtomNameOnly$Charge-";
        }
        else {
            AtomName = "$AtomNameOnly";
        }
#print "a1: $AtomName (Charge=$Charge)\n";
#print "S=$StopByError\n";
        my @a = $Rietan->ReadASFParameters($AtomName, undef, $Source, $StopByError);
        if(@a < 11) {
            AtomName = $AtomTypeList[$i]->AtomNameOnly(1);
#print "a2: $AtomName\n";
            @a = $Rietan->ReadASFParameters($AtomName, undef, $Source, $StopByError);
        }
        if(@a < 9) {
            print "Error: Can not read ASFDC for [$AtomName].\n";
            next;
        }
#print "array: ", join(',', @a), "\n";
#print "i=$i: $AtomName\n";
        this->{"ASFDCRed$i"}              = $Rietan;
        this->{"ASFDCRed$AtomName"}       = $Rietan;
        this->{"ASFDCAtomName$i"}         = $Rietan->{ASFDCAtomName};
        this->{"ASFDCAtomName$AtomName"}  = $Rietan->{ASFDCAtomName};
        AtomicNumber = $AtomTypeList[$i]->AtomicNumber();
        this->{"ASFDCAtomNameZ$i"}            = $AtomicNumber;
        this->{"ASFDCAtomNameZ$AtomNameOnly"} = $AtomicNumber;
print "ASFDC AtomName[$i]: $Rietan->{ASFDCAtomName} [Zi=$AtomicNumber]\n";
#print "   ao:$AtomNameOnly: $i\n";
    }
    $this->{ASFDCRed} = 1;
}

sub asfParameters
{
    my ($this, $iAtomType, $Source,$StopByError) = @_;
    $Source = 'Cu' if(!defined $Source);
    $StopByError = 1 if(!defined $StopByError);
    $this->ReadASF($Source,$StopByError) if(!defined $this->{ASFDCRed});

    Rietan = $this->{"ASFDCRed$iAtomType"};
#print "R=$Rietan\n";
    pP     = $Rietan->{pASFDCParameters};
#print "pP=$pP\n";
#    my ($A1,$B1,$A2,$B2,$A3,$B3,$A4,$B4,$C, $b, $df1,$df2) = @$pP;
    return @$pP;



    my ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $A5, $B5, $C, $b,
        Crdf1, $Crdf2, $Fedf1, $Fedf2, $Codf1, $Codf2, $Cudf1, $Cudf2, 
        Modf1, $Modf2, $Agdf1, $Agdf2, $Kbdf1, $Kbdf2) = @$pP;
#print "A=$A1 [$Source][$Crdf1][$Crdf2][b=$b]\n";
    if($Source eq 'Cr') {# and defined $Crdf2) {
        return ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $A5, $B5, $C, $Crdf1, $Crdf2);
    }
    elsif($Source eq 'Fe') {# and defined $Fedf2) {
        return ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $A5, $B5, $C, $Fedf1, $Fedf2);
    }
    elsif($Source eq 'Co') {# and defined $Codf2) {
        return ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $A5, $B5, $C, $Codf1, $Codf2);
    }
    elsif($Source eq 'Cu') {# and defined $Cudf2) {
        return ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $A5, $B5, $C, $Cudf1, $Cudf2);
    }
    elsif($Source eq 'Mo') {# and defined $Modf2) {
        return ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $A5, $B5, $C, $Modf1, $Modf2);
    }
    elsif($Source eq 'Ag') {# and defined $Agdf2) {
        return ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $A5, $B5, $C, $Agdf1, $Agdf2);
    }
    elsif($Source eq 'Kb') {# and defined $Kbdf2) {
        return ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $A5, $B5, $C, $Kbdf1, $Kbdf2);
    }
    return ();
}

# $s = sin Q / lambda (in nm)
sub AtomicScatteringFactor
{
    my ($this, $s) = @_;
    return $this->asf($s);
}

# $s = sin Q / lambda (in nm)
sub asfElectron
{
    my ($this, $iAtomType, $s,$StopByError) = @_;
    $StopByError = 1 if(!defined $StopByError);
    $s = 1.0e-5 if($s <= 0.0);
# the unit needs to be checked again
#    s2 = 0.01 * $s * $s; # convert nm-1 to A-1
    s2 = $s * $s;
    k  = 1.0e-2 * 8.0 * $pi * $pi / ($a0 * 1.0e9);
    Z  = $this->{"ASFDCAtomNameZ$iAtomType"};

    asfXray = $this->asf($iAtomType, $s,$StopByError);
#print "s=$s, i=$iAtomType, Z=$Z, a=$asfXray\n";
#exit;
    val = $k * ($Z - $asfXray) / $s2;
#print "va [$iAtomType]=$val = $k * ($Z - $asfXray) / $s2\n";
    return $val;
}

# $s = sin Q / lambda (in nm)
sub asf
{
    my ($this, $iAtomType, $s, $Source,$StopByError) = @_;
    $Source = 'Cu' if(!defined $Source);
    $StopByError = 1 if(!defined $StopByError);
    $this->ReadASF($Source,$StopByError) if(!defined $this->{ASFDCRed});

    Rietan = $this->{"ASFDCRed$iAtomType"};
    pP     = $Rietan->{pASFDCParameters};
    my ($A1, $B1, $A2, $B2, $A3, $B3, $A4, $B4, $A5, $B5, $C, $b, $df1,$df2) = @$pP;
    
    return 0.0 if(!defined $B5 or !defined $C);

    $df1 = 0.0 if(!defined $df1);
    $df2 = 0.0 if(!defined $df2);
    s2 = 0.01 * $s * $s; # convert nm-1 to A-1
    asf = $A1*exp(-$B1*$s2) + $A2*exp(-$B2*$s2) + $A3*exp(-$B3*$s2) + $A4*exp(-$B4*$s2) + $A5*exp(-$B5*$s2) + $C;
#    asf = $A1*exp(-$B1*$s2) + $A2*exp(-$B2*$s2) + $A3*exp(-$B3*$s2) + $A4*exp(-$B4*$s2) + $C;
#print "i=$iAtomType, s=$s [$A1,$B1] [$A2,$B2] [$A3,$B3] [$A4,$B4] [$A5,$B5] [$C]\n";
#print "cal: [$A1*exp(-$B1*$s2) + $A2*exp(-$B2*$s2) + $A3*exp(-$B3*$s2) + $A4*exp(-$B4*$s2) + $A5*exp(-$B5*$s2) + $C]\n";
#print "i=$iAtomType, s=$s, asf=$asf [A1=$A1, C=$C]\n";
    $asf = cplx($asf + $df1, $df2);
#print "i=$iAtomType, s=$s, asf=$asf [A1=$A1, C=$C] [df1=$df1, df2=$df2]\n";
    return $asf;



    my ($Crdf1,$Crdf2, $Fedf1,$Fedf2, $Codf1,$Codf2, $Cudf1,$Cudf2, 
        Modf1, $Modf2, $Agdf1,$Agdf2, $Kbdf1, $Kbdf2);

    ($A1,$B1,$A2,$B2,$A3,$B3,$A4,$B4,$C, $b,
        Crdf1,$Crdf2, $Fedf1,$Fedf2, $Codf1,$Codf2, $Cudf1,$Cudf2, 
        Modf1, $Modf2, $Agdf1,$Agdf2, $Kbdf1, $Kbdf2) = @$pP;
#    if($Source eq 'n') {
#        return $b;
#    }

    $s2 = 0.01 * $s * $s;
    $asf = $A1*exp(-$B1*$s2)+$A2*exp(-$B2*$s2)+$A3*exp(-$B3*$s2)+$A4*exp(-$B4*$s2)+$C;
    if($Source eq 'Cr' and defined $Crdf2) {
        asf = cplx($asf + $Crdf1, $Crdf2);
    }
    elsif($Source eq 'Fe' and defined $Fedf2) {
        asf = cplx($asf + $Fedf1, $Fedf2);
    }
    elsif($Source eq 'Co' and defined $Codf2) {
        asf = cplx($asf + $Codf1, $Codf2);
    }
    elsif($Source eq 'Cu' and defined $Cudf2) {
#print "d=$Cudf1,$Cudf2\n";
        asf = cplx($asf + $Cudf1, $Cudf2);
    }
    elsif($Source eq 'Mo' and defined $Modf2) {
        asf = cplx($asf + $Modf1, $Modf2);
    }
    elsif($Source eq 'Ag' and defined $Agdf2) {
        asf = cplx($asf + $Agdf1, $Agdf2);
    }
    elsif($Source eq 'Kb' and defined $Kbdf2) {
        asf = cplx($asf + $Kbdf1, $Kbdf2);
    }

    return $asf;
}

sub Fhkl
{
    my ($this, $h, $k, $l, $IsElectron,$StopByError) = @_;
    $IsElectron = 0 if(!defined $IsElectron);
    $StopByError = 1 if(!defined $StopByError);

    my @ExpandedAtomSiteList   = $this->GetCExpandedAtomSiteList();
    if(@ExpandedAtomSiteList == 0) {
        this->ExpandCoordinates();
        @ExpandedAtomSiteList = $this->GetCExpandedAtomSiteList();
    }
    nTotalExpandedAtomSite = $this->nTotalExpandedAtomSite();

    dhkl = $this->CalculateHKLInterplanarSpacing($h, $k, $l);
    s    = 1.0 / 2.0 / ($dhkl * 0.1); # in nm
#print "($h $k $l): d=$dhkl  s=$s\n";
    F  = cplx(0, 0);
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
#        AtomName  = $ExpandedAtomSiteList[$i]->AtomName();
        AtomName  = $ExpandedAtomSiteList[$i]->AtomNameOnly(1);
        iAtomType = $ExpandedAtomSiteList[$i]->iAtomType() - 1;
        my ($x,$y,$z) = $ExpandedAtomSiteList[$i]->Position(1);
        occupancy = $ExpandedAtomSiteList[$i]->Occupancy();
#print "$i: $iAtomType: $AtomName: ($x,$y,$z) occ=$occupancy [IsElectron=$IsElectron]\n";
        asf;
        if($IsElectron) {
            asf = $occupancy * $this->asfElectron($iAtomType, $s,$StopByError);
        }
        else {
            asf = $occupancy * $this->asf($iAtomType, $s,$StopByError);
        }
#print "i=$i [$h,$k,$l]: pos($x,$y,$z) o=$occupancy  asf=$asf ($AtomName:", $this->{"ASFDCAtomName$iAtomType"}, ")\n";
        phi = 2.0 * $pi * ($h * $x + $k * $y + $l * $z);
        F += $asf * exp(i * $phi);
    }
    return (Re($F), Im($F));
#    return ($Fr, $Fi);
}

sub OrthorhombicMultiplicity
{
    my ($this, $m, $h, $k, $l) = @_;
    $m *= 2 if($h != 0);
    $m *= 2 if($k != 0);
    $m *= 2 if($l != 0);
    return $m;
}

sub TetragonalMultiplicity
{
    my ($this, $m, $h, $k, $l) = @_;
    $h = abs($h);
    $k = abs($k);
    $m *= 2 if($h != 0);
    $m *= 2 if($k != 0);
    $m *= 2 if($l != 0);
    if($h != $k) {
        m *= 2;
    }
    return $m;
}

sub HexagonalMultiplicity
{
    my ($this, $m, $h, $k, $l) = @_;
    return 16 if($h != 0 and $k != 0 and $l != 0);
    return  8 if($h == $k and $l != 0);
    return 12 if($h == 0 and $k != 0 and $l != 0);
    return 12 if($h != 0 and $k == 0 and $l != 0);
    return 12 if($h != 0 and $k != 0 and $l == 0);
    return  6 if($h == $k and $l == 0);
    return  6 if($h == 0 and $k != 0 and $l == 0);
    return  6 if($h != 0 and $k == 0 and $l == 0);
    return  2 if($h == 0 and $k == 0 and $l != 0);
print "Error in CrystalObject::HexagonalMultiplicity: Invalid hkl [$h,$k,$l]\n";
    return undef;
}

sub TrigonalMultiplicity
{
    my ($this, $m, $h, $k, $l) = @_;
    return $this->HexagonalMultiplicity($m, $h, $k, $l);
}

sub RhombohedralMultiplicity
{
    my ($this, $m, $h, $k, $l) = @_;
    return $this->HexagonalMultiplicity($m, $h, $k, $l);
}

sub CubicMultiplicity
{
    my ($this, $m, $h, $k, $l) = @_;
    $h = abs($h);
    $k = abs($k);
    $l = abs($l);
    $m *= 2 if($h != 0);
    $m *= 2 if($k != 0);
    $m *= 2 if($l != 0);
    if($h != $k and $k != $l and $l != $h) {
        m *= 6;
    }
    elsif($h == $k and $k != $l) {
        m *= 3;
    }
    elsif($k == $l and $l != $h) {
        m *= 3;
    }
    elsif($l == $h and $h != $k) {
        m *= 3;
    }
    return $m;
}

sub Multiplicity
{
    my ($this, $h, $k, $l) = @_;
    ls = $this->LatticeSystem();
#print "ls: $ls\n";
    m = 1;
    if($ls eq 'cubic') {
        return $this->CubicMultiplicity($m, $h, $k, $l);
    }
    elsif($ls eq 'tetragonal') {
        return $this->TetragonalMultiplicity($m, $h, $k, $l);
    }
    elsif($ls eq 'trigonal') {
        return $this->TrigonalMultiplicity($m, $h, $k, $l);
    }
    elsif($ls eq 'hexagonal') {
        return $this->HexagonalMultiplicity($m, $h, $k, $l);
    }
    elsif($ls eq 'orthorhombic') {
        return $this->OrthorhombicMultiplicity($m, $h, $k, $l);
    }
    elsif($ls eq 'triclinic') {
        return 2;
    }
    else {
        print "Error in CrystalObject::Multiplicity: Lattice system [$ls] is not supported.\n";
        return undef;
    }
}

sub ConvertByMatrix
{
    my ($this, $x, $y, $z, $T) = @_;
    
    x1 = $T->element(1,1)*$x + $T->element(1,2)*$y + $T->element(1,3)*$z;
    y1 = $T->element(2,1)*$x + $T->element(2,2)*$y + $T->element(2,3)*$z;
    z1 = $T->element(3,1)*$x + $T->element(3,2)*$y + $T->element(3,3)*$z;
    return ($x1, $y1, $z1);
}

sub PrimitiveLatticeVectors
{
    my ($this, $UseAtomicUnit) = @_;
    k = ($UseAtomicUnit)? 1.0 / 0.529177 : 1.0;

    my ($ax,$ay,$az,$bx,$by,$bz,$cx,$cy,$cz) = $this->LatticeVectors($UseAtomicUnit);
    my ($sT, $T) = $this->GetBravaisToPrimitiveCellMatrix();
    my ($ax1, $ay1, $az1) = $this->ConvertByMatrix($ax, $ay, $az, $sT);
    my ($bx1, $by1, $bz1) = $this->ConvertByMatrix($bx, $by, $bz, $sT);
    my ($cx1, $cy1, $cz1) = $this->ConvertByMatrix($cx, $cy, $cz, $sT);
#print "vector ($ax1, $ay1, $az1, $bx1, $by1, $bz1, $cx1, $cy1, $cz1)\n";
#exit;
    return ($ax1, $ay1, $az1, $bx1, $by1, $bz1, $cx1, $cy1, $cz1);
}

sub GetCReducedAtomSiteList
{
    my ($this, $pM, $pVT, $pAtomSiteList) = @_;
    my @ExpandedAtomSiteList;
    if(!defined $pAtomSiteList) {
        @ExpandedAtomSiteList = $this->GetCExpandedAtomSiteListByOutputMode();
    }
    else {
        @ExpandedAtomSiteList = @$pAtomSiteList;
    }
    if(@$pM == 1) {
        return (1, @ExpandedAtomSiteList);
    }

    nTotalExpandedAtomSite = @ExpandedAtomSiteList;
    IsMatched = 1;
    my %iIdenticalSite;
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        next if(defined $iIdenticalSite{$i});

        atom0          = $ExpandedAtomSiteList[$i];
        atomname0      = $atom0->AtomNameOnly();
        charge0        = $atom0->Charge();
        my ($x0, $y0, $z0) = $atom0->Position(1);
#        x0             = new MyMatrixReal([[$x0], [$y0], [$z0]]);
        Vx0            = new MyMatrixReal([[$x0], [$y0], [$z0]]);
        occupancy0     = $atom0->Occupancy();
#print "$i: $atomname0 ($x0, $y0, $z0) occ=$occupancy0\n";
        IsFound = 0;
        for(my $it = 1 ; $it < @$pM ; $it++) {
            M = $pM->[$it];
            V = $pVT->[$it];
            x1 = $M * $Vx0 + $V;
#            x1 = $M * $x0 + $V;

            for(my $i1 = $i + 1 ; $i1 < $nTotalExpandedAtomSite ; $i1++) {
                atom1          = $ExpandedAtomSiteList[$i1];
                atomname1      = $atom1->AtomNameOnly();
                charge1        = $atom1->Charge();
                my ($x1, $y1, $z1) = $atom1->Position(1);
                occupancy1     = $atom1->Occupancy();
#print "   it=$it ($xt, $yt, $zt): $i1: $atomname1 ($x1, $y1, $z1): occ=$occupancy1\n";
                next if($atomname0 ne $atomname1);
                next if(abs($charge0    - $charge1   ) > 1.0e-3);
                next if(abs($occupancy0 - $occupancy1) > 1.0e-3);
#print "     candidate\n";

                dis = $this->GetNearestInterAtomicDistance(
                    x1->e(0), $x1->e(1), $x1->e(2), $x1, $y1, $z1);
                if($dis < 1.0e-3) {
#print "   identical: it=$it: $i1: $atomname1 ($x1, $y1, $z1)/($xt, $yt, $zt): $dis\n";
                    iIdenticalSite{$i1} = $iIdenticalSite{$i1} . ", $i";
                    IsFound = 1;
                    last;
                }
                else {
#print "   not same : it=$it: $i1: $atomname1 ($x1, $y1, $z1)/($xt, $yt, $zt): $dis\n";
                }
            }
        }
        IsMatched = 0 if(!$IsFound);
    }

#    foreach my $key (sort keys %iIdenticalSite) {
#print "Identical: $key - $iIdenticalSite{$key}\n";
#    }

    my @ReducedAtomSiteList;
    if($IsMatched) {
        for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
            next if($iIdenticalSite{$i} ne '');
            push(@ReducedAtomSiteList, $ExpandedAtomSiteList[$i]);
        }
    }
    else {
        @ReducedAtomSiteList = @ExpandedAtomSiteList;
    }

    return ($IsMatched, @ReducedAtomSiteList);
}

sub GetCReducedAtomSiteListByTranslation
{
    my ($this, $pTranslationalVectors, $pAtomSiteList) = @_;
    if(!defined $pTranslationalVectors) {
        pTranslationalVectors = $this->GetTranslationalVectors();
    }
if(0) {
    my (@m, @v);
    m = new MyMatrixReal('unit', 3, 3);
    for(my $i = 0 ; $i < @$pTranslationalVectors ; $i++) {
        p = $pTranslationalVectors->[$i];
        v = new MyMatrixReal([[$p->[0]], [$p->[1]], [$p->[2]]]);
        push(@m, $m);
        push(@v, $v);
    }
    return $this->GetCReducedAtomSiteList(\@m, \@v, $pAtomSiteList);
}

    my @TV = @$pTranslationalVectors;
    my @ExpandedAtomSiteList;
    if(!defined $pAtomSiteList) {
        @ExpandedAtomSiteList = $this->GetCExpandedAtomSiteListByOutputMode();
    }
    else {
        @ExpandedAtomSiteList = @$pAtomSiteList;
    }
    if(@TV == 1) {
        return (1, @ExpandedAtomSiteList);
    }
    

    nTotalExpandedAtomSite = @ExpandedAtomSiteList;
    IsMatched = 1;
    my %iIdenticalSite;
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        next if(defined $iIdenticalSite{$i});

        atom0          = $ExpandedAtomSiteList[$i];
        atomname0      = $atom0->AtomNameOnly();
        charge0        = $atom0->Charge();
        my ($x0, $y0, $z0) = $atom0->Position(1);
        occupancy0     = $atom0->Occupancy();
#print "$i: $atomname0 ($x0, $y0, $z0) occ=$occupancy0\n";
        IsFound = 0;
        for(my $it = 1 ; $it < @TV ; $it++) {
            xt = $x0 + $TV[$it][0];
            yt = $y0 + $TV[$it][1];
            zt = $z0 + $TV[$it][2];
            for(my $i1 = $i + 1 ; $i1 < $nTotalExpandedAtomSite ; $i1++) {
                atom1          = $ExpandedAtomSiteList[$i1];
                atomname1      = $atom1->AtomNameOnly();
                charge1        = $atom1->Charge();
                my ($x1, $y1, $z1) = $atom1->Position(1);
                occupancy1     = $atom1->Occupancy();
#print "   it=$it ($xt, $yt, $zt): $i1: $atomname1 ($x1, $y1, $z1): occ=$occupancy1\n";
                next if($atomname0 ne $atomname1);
                next if(abs($charge0    - $charge1   ) > 1.0e-3);
                next if(abs($occupancy0 - $occupancy1) > 1.0e-3);
#print "     candidate\n";

                dis = $this->GetNearestInterAtomicDistance($xt, $yt, $zt, $x1, $y1, $z1);
                if($dis < 1.0e-3) {
#print "   identical: it=$it: $i1: $atomname1 ($x1, $y1, $z1)/($xt, $yt, $zt): $dis\n";
                    iIdenticalSite{$i1} = $iIdenticalSite{$i1} . ", $i";
                    IsFound = 1;
                    last;
                }
                else {
#print "   not same : it=$it: $i1: $atomname1 ($x1, $y1, $z1)/($xt, $yt, $zt): $dis\n";
                }
            }
        }
        IsMatched = 0 if(!$IsFound);
    }

#    foreach my $key (sort keys %iIdenticalSite) {
#print "Identical: $key - $iIdenticalSite{$key}\n";
#    }

    my @ReducedAtomSiteList;
    if($IsMatched) {
        for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
            next if($iIdenticalSite{$i} ne '');
            push(@ReducedAtomSiteList, $ExpandedAtomSiteList[$i]);
        }
    }
    else {
        @ReducedAtomSiteList = @ExpandedAtomSiteList;
    }

    return ($IsMatched, @ReducedAtomSiteList);
}

sub GetCReducedAtomSiteListByInversion
{
    my ($this, $pAtomSiteList) = @_;
if(0) {
    m = new MyMatrixReal([[-1, 0, 0], [0, -1, 0], [0, 0, -1]]);
    v = new MyMatrixReal([[0], [0], [0]]);
    return $this->GetCReducedAtomSiteList([$m], [$v], $pAtomSiteList);
}
    ($this, $pAtomSiteList) = @_;
    my @ExpandedAtomSiteList;
    if(!defined $pAtomSiteList) {
        @ExpandedAtomSiteList = $this->GetCExpandedAtomSiteListByOutputMode();
    }
    else {
        @ExpandedAtomSiteList = @$pAtomSiteList;
    }

    nTotalExpandedAtomSite = @ExpandedAtomSiteList;
    IsMatched = 1;
    my %iIdenticalSite;
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        next if(defined $iIdenticalSite{$i});

        atom0          = $ExpandedAtomSiteList[$i];
        atomname0      = $atom0->AtomNameOnly();
        charge0        = $atom0->Charge();
        my ($x0, $y0, $z0) = $atom0->Position(1);
        occupancy0     = $atom0->Occupancy();
#print "$i: $atomname0 ($x0, $y0, $z0) occ=$occupancy0\n";
        IsFound = 0;
        my ($xt, $yt, $zt) = (Utils::Reduce01(-$x0), Utils::Reduce01(-$y0), Utils::Reduce01(-$z0));
        for(my $i1 = $i ; $i1 < $nTotalExpandedAtomSite ; $i1++) {
            atom1          = $ExpandedAtomSiteList[$i1];
            atomname1      = $atom1->AtomNameOnly();
            charge1        = $atom1->Charge();
            my ($x1, $y1, $z1) = $atom1->Position(1);
            occupancy1     = $atom1->Occupancy();
#print "   ($xt, $yt, $zt): $i1: $atomname1 ($x1, $y1, $z1): occ=$occupancy1\n";
            next if($atomname0 ne $atomname1);
            next if(abs($charge0    - $charge1   ) > 1.0e-3);
            next if(abs($occupancy0 - $occupancy1) > 1.0e-3);

            dis = $this->GetNearestInterAtomicDistance($xt, $yt, $zt, $x1, $y1, $z1);
            if($dis < 1.0e-3) {
#print "   identical: $i1: $atomname1 ($x1, $y1, $z1)/($xt, $yt, $zt): $dis\n";
                iIdenticalSite{$i1} = $iIdenticalSite{$i1} . ", $i";
                IsFound = 1;
                last;
            }
            else {
#print "   not same : $i1: $atomname1 ($x1, $y1, $z1)/($xt, $yt, $zt): $dis\n";
            }
        }
        IsMatched = 0 if(!$IsFound);
    }

#    foreach my $key (sort keys %iIdenticalSite) {
#print "Identical: $key - $iIdenticalSite{$key}\n";
#    }

    my @ReducedAtomSiteList;
    if($IsMatched) {
        for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
            next if($iIdenticalSite{$i} ne '');
            push(@ReducedAtomSiteList, $ExpandedAtomSiteList[$i]);
        }
    }
    else {
        @ReducedAtomSiteList = @ExpandedAtomSiteList;
    }

    return ($IsMatched, @ReducedAtomSiteList);
}

sub AnalyzeaKProduct
{
    my ($this, $aKProduct) = @_;

    my ($a,$b,$c,$alpha,$beta,$gamma) = $this->LatticeParametersByOutputMode(0);
    my ($nx, $ny, $nz);
    an = $aKProduct * 10.0;
    $nx = int($an / $a) + 1;
    $ny = int($an / $b) + 1;
    $nz = int($an / $c) + 1;

    return ($nx, $ny, $nz);
}

sub GetTranslationalVectors
{
    my ($this) = @_;

    SPG        = $this->GetCSpaceGroup();
    SPGName    = $SPG->SPGName();

    my @VT = ([  0,   0,   0]);
    if($SPGName =~ /^\s*F/i) {
        @VT = (
            [  0,   0,   0],
            [  0, 1/2, 1/2],
            [1/2,   0, 1/2],
            [1/2, 1/2,   0],
            );
    }
    elsif($SPGName =~ /^\s*I/i) {
        @VT = (
            [  0,   0,   0],
            [1/2, 1/2, 1/2],
            );
    }
    elsif($SPGName =~ /^\s*A/i) {
        @VT = (
            [  0,   0,   0],
            [  0, 1/2, 1/2],
            );
    }
    elsif($SPGName =~ /^\s*B/i) {
        @VT = (
            [  0,   0,   0],
            [1/2,   0, 1/2],
            );
    }
    elsif($SPGName =~ /^\s*C/i) {
        @VT = (
            [  0,   0,   0],
            [1/2, 1/2,   0],
            );
    }
    return \@VT;
}

sub IsFaceCenteredCell
{
    my ($this, $pAS, $eps) = @_;

    my @m = (
        [0.0, 0.5, 0.5],
        [0.5, 0.0, 0.5],
        [0.5, 0.5, 0.0],
        );

    return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub IsBodyCenteredCell
{
    my ($this, $pAS, $eps) = @_;

    my @m = (
        [0.5, 0.5, 0.5]
        );

    return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub IsACenteredCell
{
    my ($this, $pAS, $eps) = @_;

    my @m = (
        [0.0, 0.5, 0.5]
        );

    return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub IsBCenteredCell
{
    my ($this, $pAS, $eps) = @_;

    my @m = (
        [0.5, 0.0, 0.5]
        );

    return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub IsCCenteredCell
{
    my ($this, $pAS, $eps) = @_;

    my @m = (
        [0.5, 0.5, 0.0]
        );

    return $this->IsBravaisCellByMatrix($pAS, $eps, \@m);
}

sub GuessBravaisLattice
{
    my ($this, $tollatt, $tolangle, $EPSCoord) = @_;

    my ($a, $b, $c, $alpha, $beta, $gamma) = $this->LatticeParameters();
    SPG     = $this->GetCSpaceGroup();
    SPGName = $SPG->SPGName();
    ls      = lc $this->LatticeSystem();
#print "SPG=$SPGName  LS=$ls\n";

    LatticeSystem = ($ls eq '')? $SPG->SetLatticeSystem(undef, 1, $tollatt, $tolangle) : $ls;

    BravaisLattice;
    if($SPGName =~ /^\s*F/i) {
        BravaisLattice = 'F';
    }
    elsif($SPGName =~ /^\s*I/i) {
        BravaisLattice = 'I';
    }
    elsif($SPGName =~ /^\s*A/i) {
        BravaisLattice = 'A';
    }
    elsif($SPGName =~ /^\s*B/i) {
        BravaisLattice = 'B';
    }
    elsif($SPGName =~ /^\s*C/i) {
        BravaisLattice = 'C';
    }
#print "SPG=$SPGName  BL=$BravaisLattice\n";

    if($BravaisLattice eq '') {
        my @as = $this->GetCExpandedAtomSiteListByOutputMode();
        if($this->IsFaceCenteredCell(\@as, $EPSCoord)) {
            BravaisLattice = 'F';
        }
        elsif($this->IsBodyCenteredCell(\@as, $EPSCoord)) {
            BravaisLattice = 'I';
        }
        elsif($this->IsACenteredCell(\@as, $EPSCoord)) {
            BravaisLattice = 'A';
        }
        elsif($this->IsBCenteredCell(\@as, $EPSCoord)) {
            BravaisLattice = 'B';
        }
        elsif($this->IsCCenteredCell(\@as, $EPSCoord)) {
            BravaisLattice = 'C';
        }
        else {
            BravaisLattice = 'P';
        }
    }

    return ($LatticeSystem, $BravaisLattice);
}

sub GetLatticeConversionMatrix
{
    my ($this,  $PresetRule, $DirectionSelect, $ParametersSelect, $IsPrint) = @_;
    $IsPrint = 1 if(!defined $IsPrint);

    sT  = new Math::MatrixReal(3, 3);
    T   = new Math::MatrixReal(3, 3);
    tRT = new Math::MatrixReal(3, 3);
    RT  = new Math::MatrixReal(3, 3);
    tR  = new Math::MatrixReal(3, 3);

    if($PresetRule eq 'No' or $PresetRule eq 'P1') {
        sT = Math::MatrixReal->new_from_cols( 
                [ [ 1,  0,  0], 
                  [ 0,  1,  0],
                  [ 0,  0,  1] ] );
        sT->transpose($sT);
    }
    elsif($PresetRule eq 'RhombHex') {
        sT = Math::MatrixReal->new_from_cols( 
                [ [ 1, -1,  0], 
                  [ 0,  1, -1],
                  [ 1,  1,  1] ] );
        sT->transpose($sT);
    }
    elsif($PresetRule eq 'HexRhomb') {
        sT = Math::MatrixReal->new_from_cols( 
                [ [ 2/3, 1/3, 1/3],
                  [-1/3, 1/3, 1/3],
                  [-1/3,-2/3, 1/3] ]);
        sT->transpose($sT);
    }
    elsif($PresetRule eq 'HexOrtho') {
        sT = Math::MatrixReal->new_from_cols( 
                [ [ 1, 0, 0], 
                  [ 1, 2, 0],
                  [ 0, 0, 1] ] );
        sT->transpose($sT);
    }
    elsif($PresetRule eq 'FCCPrim' or $PresetRule eq 'PrimFCC') {
        sT = Math::MatrixReal->new_from_cols( 
                [ [ 0.5, 0.5,   0], 
                  [ 0,   0.5, 0.5],
                  [ 0.5,   0, 0.5] ] );
        if($PresetRule eq 'PrimFCC') {
            sT = $sT->inverse();
        }
        sT->transpose($sT);
    }
    elsif($PresetRule eq 'BCCPrim' or $PresetRule eq 'PrimBCC') {
        sT = Math::MatrixReal->new_from_cols( 
                [ [-0.5, 0.5, 0.5], 
                  [ 0.5,-0.5, 0.5],
                  [ 0.5, 0.5,-0.5] ] );
        if($PresetRule eq 'PrimBCC') {
            sT = $sT->inverse();
        }
        sT->transpose($sT);
    }
    elsif($PresetRule eq 'ACenterPrim') {
        sT = Math::MatrixReal->new_from_cols( 
                [ [ 1.0, 0.0, 0.0], 
                  [ 0.0, 0.5, 0.5],
                  [ 0.0,-0.5, 0.5] ] );
        sT->transpose($sT);
    }
    elsif($PresetRule eq 'BCenterPrim') {
        sT = Math::MatrixReal->new_from_cols( 
                [ [ 0.5, 0.0, 0.5], 
                  [ 0.0, 1.0, 0.0],
                  [-0.5, 0.0, 0.5] ] );
        sT->transpose($sT);
    }
    elsif($PresetRule eq 'CCenterPrim') {
        sT = Math::MatrixReal->new_from_cols( 
                [ [ 0.5, 0.5, 0.0], 
                  [-0.5, 0.5, 0.0],
                  [ 0.0, 0.0, 1.0] ] );
        sT->transpose($sT);
    }
    else {
die "Error: Invalid conversion rule [$PresetRule].\n";

        print "  Rule is specified by Matrix:\n";
        sT->assign(1, 1, eval(&GetArg('t11')));
        sT->assign(1, 2, eval(&GetArg('t12')));
        sT->assign(1, 3, eval(&GetArg('t13')));
        sT->assign(2, 1, eval(&GetArg('t21')));
        sT->assign(2, 2, eval(&GetArg('t22')));
        sT->assign(2, 3, eval(&GetArg('t23')));
        sT->assign(3, 1, eval(&GetArg('t31')));
        sT->assign(3, 2, eval(&GetArg('t32')));
        sT->assign(3, 3, eval(&GetArg('t33')));
    }

print "Conversion rule:\n" if($IsPrint);
    IsChosen = 0;
    if($ParametersSelect eq 'ai') {
        if($DirectionSelect eq 'OriginalToConverted') {
            print "Real space vectors: (a'i) = (tij)(ai)\n" if($IsPrint);
            T = $sT;
            IsChosen = 1;
        }
        elsif($DirectionSelect eq 'ConvertedlToOriginal') {
            print "Real space vectors: (ai) = (tij)(a'i)\n" if($IsPrint);
            T = $sT->inverse();
            IsChosen = 1;
        }
    }
    elsif($ParametersSelect eq 'xyz') {
        if($DirectionSelect eq 'OriginalToConverted') {
            print "Real space coordinates: (x'i) = (tij)(xi)\n" if($IsPrint);
            T->transpose($sT);
            T = $T->inverse();
            IsChosen = 1;
        }
        elsif($DirectionSelect eq 'ConvertedlToOriginal') {
            print "Real space vectors: (xi) = (tij)(x'i)\n" if($IsPrint);
            T->transpose($sT);
            IsChosen = 1;
        }
    }
    elsif($ParametersSelect eq 'a*i') {
        if($DirectionSelect eq 'OriginalToConverted') {
            print "Reciprocal space vectors: (a'*i) = (tij)(a*i)\n" if($IsPrint);
            T->transpose($sT);
            T = $T->inverse();
            IsChosen = 1;
        }
        elsif($DirectionSelect eq 'ConvertedlToOriginal') {
            print "Reciprocal space vectors: (a*i) = (tij)(a'*i)\n" if($IsPrint);
            T->transpose($sT);
            IsChosen = 1;
        }
    }
    elsif($ParametersSelect eq 'hkl') {
        if($DirectionSelect eq 'OriginalToConverted') {
            print "Reciprocal space coordinates: (h'i) = (tij)(hi)\n" if($IsPrint);
            T = $sT;
            IsChosen = 1;
        }
        elsif($DirectionSelect eq 'ConvertedlToOriginal') {
            print "Reciprocal space coordinates: (hi) = (tij)(h'i)\n" if($IsPrint);
            T = $sT->inverse();
            IsChosen = 1;
        }
    }
    unless($IsChosen) {
print "Error in CrystalObject::GetLatticeConversionMatrix: Selected option {Parameters: $ParametersSelect} in {$DirectionSelect} direction is not supported.\n";
        return 1;
    }

#実空間座標(x,y,z)の変換行列
    $tRT->transpose($T);
    $tRT = $tRT->inverse();

#逆空間ベクトル(a*i)の変換行列: $tRT
    $RT->transpose($T);
    $RT = $RT->inverse();

#逆空間座標(h k l)の変換行列: $T
    $tR = $T;

    return ($sT, $T, $tRT, $RT, $tR);
}

sub GetPrimitiveCrystal
{
    my ($this, $LatticeParameterOnly, $CheckConsistency, $IsPrint) = @_;

    T = new Math::MatrixReal(3, 3);
    matrix = $this->GetMatrixConventionalToPrimitiveCell(0);
    $T = Math::MatrixReal->new_from_cols($matrix);
    $T->transpose($T);
    return $this->ConvertLattice($T, $LatticeParameterOnly, $CheckConsistency, $IsPrint);
}

sub GetMatrixConventionalToPrimitiveCell
{
    my ($this, $ReferToConventionalCell) = @_;
    SPGName = ($ReferToConventionalCell and $this->{ConventionalSPGName})? 
                this->{ConventionalSPGName} : $this->SPGName();
#print "2********[$this->{ConventionalSPGName}][$SPGName]\n";

    if($SPGName =~ /^\s*F/i) {
        return [ [ 0.5, 0.5,   0], 
             [ 0,   0.5, 0.5],
             [ 0.5,   0, 0.5] ]
    }
    elsif($SPGName =~ /^\s*I/i) {
        return [ [-0.5, 0.5, 0.5], 
             [ 0.5,-0.5, 0.5],
             [ 0.5, 0.5,-0.5] ];
    }
    elsif($SPGName =~ /^\s*A/i) {
        return [ [ 1.0, 0.0, 0.0], 
             [ 0.0, 0.5, 0.5],
             [ 0.0,-0.5, 0.5] ];
    }
    elsif($SPGName =~ /^\s*B/i) {
        return [ [ 0.5, 0.0, 0.5], 
             [ 0.0, 1.0, 0.0],
             [-0.5, 0.0, 0.5] ];
    }
    elsif($SPGName =~ /^\s*C/i) {
        return [ [ 0.5, 0.5, 0.0], 
             [-0.5, 0.5, 0.0],
             [ 0.0, 0.0, 1.0] ];
    }
    return [ [ 1.0, 0.0, 0.0], 
         [ 0.0, 1.0, 0.0],
         [ 0.0, 0.0, 1.0] ];
}

sub GetMatrixConventionalToPrimitiveReciprocalCell
{
    my ($this) = @_;
    SPGName = $this->SPGName();

    if($SPGName =~ /^\s*F/i) {
        return [ [-1.0, 1.0, 1.0], 
             [ 1.0,-1.0, 1.0],
             [ 1.0, 1.0,-1.0] ];
    }
    elsif($SPGName =~ /^\s*I/i) {
        return [ [ 1.0, 1.0,   0], 
             [ 0,   1.0, 1.0],
             [ 1.0,   0, 1.0] ]
    }
    elsif($SPGName =~ /^\s*A/i) {
        return [ [ 1.0, 0.0, 0.0], 
             [ 0.0, 1.0, 1.0],
             [ 0.0,-1.0, 1.0] ];
    }
    elsif($SPGName =~ /^\s*B/i) {
        return [ [ 1.0, 0.0, 1.0], 
             [ 0.0, 1.0, 0.0],
             [-1.0, 0.0, 1.0] ];
    }
    elsif($SPGName =~ /^\s*C/i) {
        return [ [ 1.0, 1.0, 0.0], 
             [-1.0, 1.0, 0.0],
             [ 0.0, 0.0, 1.0] ];
    }
    return [ [ 1.0, 0.0, 0.0], 
         [ 0.0, 1.0, 0.0],
         [ 0.0, 0.0, 1.0] ];
}


sub IsBravaisCellByMatrix
{
    my ($this, $pAS, $eps, $pm) = @_;
    $eps = 0.01 if(!defined $eps);

    my @Checked;
    for(my $i = 0 ; $i < @$pAS ; $i++) {
        next if($Checked[$i]);

        site1 = $pAS->[$i];
        atom1          = $site1->AtomNameOnly();
        my ($x1, $y1, $z1) = $site1->Position(1);
        occ1           = $site1->Occupancy();
#Utils::Reduce01($x)

        for(my $im = $0 ; $im < @$pm ; $im++) {
            xm = $x1 + $pm->[$im][0];
            ym = $y1 + $pm->[$im][1];
            zm = $z1 + $pm->[$im][2];

            IsFound = 0;
            for(my $j = $i+1 ; $j < @$pAS ; $j++) {
                site2 = $pAS->[$j];
                atom2          = $site2->AtomNameOnly();
                my ($x2, $y2, $z2) = $site2->Position(1);
                occ2           = $site2->Occupancy();

                next if($atom1 ne $atom2);
                dis = $this->GetNearestInterAtomicDistance(
                        xm, $ym, $zm, $x2, $y2, $z2);
                if($dis < $eps) {
                    IsFound     = 1;
                    Checked[$j] = 1;
                    last;
                }
            }

            return 0 if(!$IsFound);
        }
    }

    return 1;
}

sub GetBravaisToPrimitiveCellMatrix
{
    my ($this) = @_;

    SPG        = $this->GetCSpaceGroup();
    SPGName    = $SPG->SPGName();

    sT = new Math::MatrixReal(3, 3);
    T;
    if($SPGName =~ /^\s*F/i) {
        sT = Math::MatrixReal->new_from_cols( 
            [ [ 0.5, 0.5,   0],
              [ 0,   0.5, 0.5],
              [ 0.5,   0, 0.5] ] );
    }
    elsif($SPGName =~ /^\s*I/i) {
        sT = Math::MatrixReal->new_from_cols( 
            [ [-0.5, 0.5, 0.5],
              [ 0.5,-0.5, 0.5],
              [ 0.5, 0.5,-0.5] ] );
    }
    elsif($SPGName =~ /^\s*A/i) {
        sT = Math::MatrixReal->new_from_cols( 
            [ [ 1.0, 0.0, 0.0],
              [ 0.0, 0.5, 0.5],
              [ 0.0,-0.5, 0.5] ] );
    }
    elsif($SPGName =~ /^\s*B/i) {
        sT = Math::MatrixReal->new_from_cols( 
            [ [ 0.5, 0.0, 0.5],
              [ 0.0, 1.0, 0.0],
              [-0.5, 0.0, 0.5] ] );
    }
    elsif($SPGName =~ /^\s*C/i) {
        sT = Math::MatrixReal->new_from_cols( 
            [ [ 0.5, 0.5, 0.0],
              [-0.5, 0.5, 0.0],
              [ 0.0, 0.0, 1.0] ] );
    }
    $T = $sT;
    $T->transpose($T);
    
    return ($sT, $T);
}

sub GetPrimitiveCell
{
    my ($this, $IsPrint) = @_;

#実空間べクトル(ai)の変換行列
    my ($sT, $T) = $this->GetBravaisToPrimitiveCellMatrix();
if($IsPrint) {
    print "Conversion matrix for real space vector: (a'i) = (Tij)(ai)\n";
    printf "   |%10.4f %10.4f %10.4f|\n", $T->element(1,1), $T->element(1,2), $T->element(1,3);
    printf "   |%10.4f %10.4f %10.4f|\n", $T->element(2,1), $T->element(2,2), $T->element(2,3);
    printf "   |%10.4f %10.4f %10.4f|\n", $T->element(3,1), $T->element(3,2), $T->element(3,3);
}
    return $this->ConvertLattice($T, 0, 1);
}

sub FindTranslations
{
    my ($this, $OverlapCriteriaRadius, $IsPrint) = @_;

    SPG = $this->GetCSpaceGroup();

    my @TranslationArrays = $SPG->GetTranslationArrays();

    my (@hit, $pID, @ID);
    for(my $i = 0 ; $i < @TranslationArrays ; $i++) {
        ($hit[$i], $pID) = $this->CheckATransration(
                    TranslationArrays[$i], $OverlapCriteriaRadius, $IsPrint);
        if($hit[$i]) {
#            print "** Translation [TranslationArrays[$i]{Name}] is found\n";
            for(my $j = 0 ; $j < @$pID ; $j++) {
                ID[$j] = 1 if($pID->[$j]);
            }
        }
    }

    if($hit[0] and $hit[1] and $hit[2]) {
        print "** Translation [F] is found\n" if($IsPrint);
        return ("F", \@ID);
    }
    for(my $i = 0 ; $i < @TranslationArrays ; $i++) {
        if($hit[$i]) {
            print "** Translation [$TranslationArrays[$i]{Name}] is found\n" if($IsPrint);
            return ($TranslationArrays[$i]{Name}, \@ID);
        }
    }
    return ('', \@ID);
}

sub CheckATransration
{
    my ($this, $ppT, $OverlapCriteriaRadius, $IsPrint) = @_;
#$IsPrint=1;

    pT = $ppT->{Matrix};

    my @SiteList = $this->GetCExpandedAtomSiteList();
    nSite = @SiteList;

    my @SiteId;
    for(my $i = 0 ; $i < $nSite ; $i++) {
        next if($SiteId[$i]);

        p         = $SiteList[$i];
        label     = $p->Label();
        atomname  = $p->AtomName();
        my ($x,$y,$z) = $p->Position(1);
        occupancy = $p->Occupancy();
print "   $i: $label ($atomname): ($x, $y, $z) [$occupancy]\n" if($IsPrint);

        xs = $x + $pT->[0];
        ys = $y + $pT->[1];
        zs = $z + $pT->[2];
        id2 = $this->FindIdenticalSite($atomname, $xs, $ys, $zs, $OverlapCriteriaRadius, $i);

        if(defined $id2) {
print " Tr Op ($pT->[0], $pT->[1], $pT->[2]): ($xs,$ys,$zs): atom site $i is equivalent to site $id2\n" if($IsPrint);
            SiteId[$id2] = 1 if($i < $id2);
            next;
        }
        else {
print " Tr Op ($pT->[0], $pT->[1], $pT->[2]): ($xs,$ys,$zs): Equivalent site was not found for atom site $i\n" if($IsPrint);
            return (0);
        }
    }
    
    return (1, \@SiteId);
}

sub FindSiteSymmetries
{
    my ($this, $xc, $yc, $zc, $OverlapCriteriaRadius, $IsPrint) = @_;

    my @MArrays = PointGroup->new()->GetSymmetryOperationArraysInCartesian();
#print "     Check site symmetry for $label [$type] at ($xc, $yc, $zc)...\n";

    my ($rxc, $ryc, $rzc) = $this->CalCartesianCoordinate($xc, $yc, $zc);

    my @AtomSiteList           = $this->GetCAsymmetricAtomSiteList();
    nAsymmetricAtomSite    = @AtomSiteList;
    my @ExpandedAtomSiteList   = $this->GetCExpandedAtomSiteList();
    nTotalExpandedAtomSite = $this->nTotalExpandedAtomSite();
#    my @nMultiplicityExpandedAtomSiteList = $Crystal->GetCnMultiplicityExpandedAtomSiteList();

    my (@hit);
    for(my $is = 0 ; $is < @MArrays ; $is++) {
        pM   = $MArrays[$is]->{Matrix};
        Name = $MArrays[$is]->{Name};
        pF   = $MArrays[$is]->{Freedom};

        hit[$is] = 1;
        for(my $i = 0 ; $i < $nAsymmetricAtomSite ; $i++) {
            p         = $AtomSiteList[$i];
            label     = $p->Label();
            type      = $p->AtomName();
            my ($x,$y,$z) = $p->Position(1);
            occupancy = $p->Occupancy();

            x0 = $x - $xc;
            y0 = $y - $yc;
            z0 = $z - $zc;
            my ($rx0, $ry0, $rz0) = $this->CalCartesianCoordinate($x0, $y0, $z0);

#            xs = $pM->[0][0] * $x0 + $pM->[0][1] * $y0 + $pM->[0][2] * $z0 + $pM->[0][3] + $xc;
#            ys = $pM->[1][0] * $x0 + $pM->[1][1] * $y0 + $pM->[1][2] * $z0 + $pM->[1][3] + $yc;
#            zs = $pM->[2][0] * $x0 + $pM->[2][1] * $y0 + $pM->[2][2] * $z0 + $pM->[2][3] + $zc;
            rxs = $pM->[0][0] * $rx0 + $pM->[0][1] * $ry0 + $pM->[0][2] * $rz0 + $pM->[0][3] + $rxc;
            rys = $pM->[1][0] * $rx0 + $pM->[1][1] * $ry0 + $pM->[1][2] * $rz0 + $pM->[1][3] + $ryc;
            rzs = $pM->[2][0] * $rx0 + $pM->[2][1] * $ry0 + $pM->[2][2] * $rz0 + $pM->[2][3] + $rzc;
            my ($xs, $ys, $zs) = $this->CalFractionalCoordinate($rxs, $rys, $rzs);
#if($Name =~ /3z/) {
#print "(x,y,z)=($x,$y,$z)  0=($x0,$y0,$z0)\n";
#print "zs = $xs = $pM->[0][0] * $x0 + $pM->[0][1] * $y0 + $pM->[0][2] * $z0 + $pM->[0][3] + $xc\n";
#print "ys = $ys = $pM->[1][0] * $x0 + $pM->[1][1] * $y0 + $pM->[1][2] * $z0 + $pM->[1][3] + $yc\n";
#print "zs = $zs = $pM->[2][0] * $x0 + $pM->[2][1] * $y0 + $pM->[2][2] * $z0 + $pM->[2][3] + $zc\n";
#}

            id2 = $this->FindIdenticalSite($type, $xs, $ys, $zs, $OverlapCriteriaRadius);#, $i);
#$IsPrint=1;
            if(defined $id2) {
printf " Sym Op[$is] ($Name): ($xs,$ys,$zs): Atom ($x,$y,$z) is equivalent to site $id2\n",
                    xs, $ys, $zs, $x, $y, $z if($IsPrint);
                next;
            }
            else {
printf " Sym Op[$is] ($Name): (%7.4f,%7.4f,%7.4f): Equivalent site of (%7.4f,%7.4f,%7.4f) was not found\n",
                    xs, $ys, $zs, $x, $y, $z if($IsPrint);
                hit[$is] = 0;
                last;
            }
        }
    }

    my ($PG) = PointGroup->new()->FindPointGroup(\@hit, \@MArrays);
print "      Point Group: $PG\n";

print "      Matched symmetries: ";
    for(my $i = 0 ; $i < @MArrays ; $i++) {
        next if(!$hit[$i]);

        Name = $MArrays[$i]->{Name};
print "$Name ";
    }
print "\n";

if(0) {
print "      Incompatible symmetries: ";
    for(my $i = 0 ; $i < @MArrays ; $i++) {
        next if($hit[$i]);

        Name = $MArrays[$i]->{Name};
print "$Name ";
    }
print "\n";
}

    return (1, \@hit);
}

sub CheckSPG
{
    my ($this, $TSym, $pID, $iSPG, $iSet, $OverlapCriteriaRadius, $IsPrint) = @_;


    my ($a,$b,$c,$alpha,$beta,$gamma) = $this->LatticeParameters();
#    $Crystal->ExpandCoordinates();
    SPG     = $this->GetCSpaceGroup();
#    SPGName = $SPG->SPGName();
#    iSPG    = $SPG->iSPG();
    LatticeSystem = $SPG->LatticeSystem();
#    print "Space Group: $SPGName ($iSPG)\n";
#    print "Lattice System: $LatticeSystem\n";
    my @ExpandedAtomSiteList              = $this->GetCExpandedAtomSiteList();
    nTotalExpandedAtomSite            = $this->nTotalExpandedAtomSite();
    my @nMultiplicityExpandedAtomSiteList = $this->GetCnMultiplicityExpandedAtomSiteList();
#    print "nTotalExpandedAtomSite: $nTotalExpandedAtomSite\n";

    r              = new RietanFP;
    SPG2           = $r->ReadSpaceGroup($iSPG, $iSet);
#print "i=$iSPG, $iSet SPG2=$SPG2\n";
    return (undef) if(!$SPG2);
    SPGName        = $SPG2->SPGName();
    if($TSym ne '' and $SPGName !~ /^$TSym/i) {
        return ($SPG2, 0, "[$SPGName] in SPG is not compatible with translation symmetry [$TSym]");
    }
    LatticeSystem2 = $SPG2->LatticeSystem();

#print "LS: [$LatticeSystem][$LatticeSystem2]\n";
    if(!$SPG2->IsCompatibleLatticeSystem($LatticeSystem, $LatticeSystem2)) {
        return ($SPG2, 0, "[$LatticeSystem2] in SPG is not compatible with [$LatticeSystem]");
    }

#非対称構造中のパラメータ
    nTranslation       = $SPG2->nTranslation();
    nSymmetryOperation = $SPG2->nSymmetryOperation();
    nTotalSymmetry     = $nTranslation * $nSymmetryOperation;
print "nTotalSymmetry: $nTotalSymmetry = $nSymmetryOperation * $nTranslation\n" if($IsPrint);

    my @SiteId;
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        next if($pID->[$i]);
        next if($SiteId[$i]);

        label     = $ExpandedAtomSiteList[$i]->Label();
        atomname  = $ExpandedAtomSiteList[$i]->AtomName();
        my ($x,$y,$z) = $ExpandedAtomSiteList[$i]->Position(1);
        occupancy = $ExpandedAtomSiteList[$i]->Occupancy();
        id        = $ExpandedAtomSiteList[$i]->IdAsymmetricAtomSite();
        mult      = $nMultiplicityExpandedAtomSiteList[$id-1];
print "   $i: [$id]$label ($atomname): ($x, $y, $z) [$occupancy][$mult]\n" if($IsPrint);

        Found = 0;
        for(my $is = 0 ; $is < $nTotalSymmetry ; $is++) {
            my ($xs, $ys, $zs) = $SPG2->DoSymmetryOperation($is, $x, $y, $z, 1);
#print "        ($is) xs,ys,zs: $xs,$ys,$zs<br>\n";
#            my ($fxs, $fys, $fzs) = $SPG->DoVelocitySymmetryOperation($is, $fx, $fy, $fz, 1);
#            my ($vxs, $vys, $vzs) = $SPG->DoVelocitySymmetryOperation($is, $vx, $vy, $vz, 1);

#            id2 = $this->FindIdenticalSite($atomname, $xs, $ys, $zs, $OverlapCriteriaRadius);
            id2 = $this->FindIdenticalSite($atomname, $xs, $ys, $zs, $OverlapCriteriaRadius, $i);
            if(defined $id2) {
print " Sym Op $is ($xs,$ys,$zs): atom site $i is equivalent to site $id2\n" if($IsPrint);
                SiteId[$id2] = 1 if($i < $id2);
                Found = 1;
                next;
            }
            else {
print " Sym Op $is ($xs,$ys,$zs): Equivalent site was not found for atom site $i\n" if($IsPrint);
                return ($SPG2, 0, "No equivalent site found for site $i");
            }
        }
        if(!$Found) {
print "   .Equivalent site was not found for atom site $i\n" if($IsPrint);
            return ($SPG2, 0, "No equivalent site found for site $i");
        }
    }

    my @AsymSites;
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        next if(defined $SiteId[$i]);
        push(@AsymSites, $ExpandedAtomSiteList[$i]);
    }

    return ($SPG2, 1, "OK", \@SiteId, \@AsymSites);
}

sub FindOrigin
{
    my ($this, $xc, $yc, $zc, $OverlapCriteriaRadius, 
        ShowAllCoordination, $RCoordMax, $IsPrint) = @_;

    Rmax = $RCoordMax;
    kRth = 0.1;

    print "\n** Find origin from initial center ($xc, $yc, $zc)\n";

    my @MArrays = PointGroup->new()->GetSymmetryOperationArraysInCartesian();

    my ($la, $lb, $lc, $alpha, $beta, $gamma) = $this->LatticeParameters();
    my @AtomSiteList = $this->GetCExpandedAtomSiteList();
    nTotalSite   = $this->nTotalExpandedAtomSite();

if(0) {
    my (@site, %pSiteArray);
    for(my $i = 0 ; $i < $nTotalSite ; $i++) {
        site      = $AtomSiteList[$i];
        atomname  = $site->AtomNameOnly();
        my ($x,$y,$z) = $site->Position(1);
        occupancy = $site->Occupancy();
        ID        = sprintf("%s%04d", $atomname, int($occupancy*1000.0+1e-3));

        my %h = (
            ID          => $ID,
            'name'      => $atomname,
            'x'         => $x,
            'y'         => $y,
            'z'         => $z,
            'occupancy' => $occupancy,
            );

        pSiteArray{$ID} = [] if(!$pSiteArray{$ID});
        p = $pSiteArray{$ID};
        push(@$p, \%h);
#        print "   $i: $ID ($x, $y, $z)\n";
    }

    foreach my $key (sort keys %pSiteArray) {
        p = $pSiteArray{$key};
        print "$key: \n";
        for(my $i = 0 ; $i < @$p ; $i++) {
            print "   ($p->[$i]{x}, $p->[$i]{y}, $p->[$i]{z}) occ=$p->[$i]{occupancy}\n";
        }
    }
}

    my %pRArray;

    nxmax = int($Rmax / $la) + 1;
    nymax = int($Rmax / $lb) + 1;
    nzmax = int($Rmax / $lc) + 1;
    for(my $nz = -$nzmax ; $nz <= $nzmax ; $nz++) {
    for(my $ny = -$nymax ; $ny <= $nymax ; $ny++) {
    for(my $nx = -$nxmax ; $nx <= $nxmax ; $nx++) {
        for(my $i = 0 ; $i < $nTotalSite ; $i++) {
            site      = $AtomSiteList[$i];
            atomname  = $site->AtomNameOnly();
            my ($x,$y,$z) = $site->Position(1);
            occupancy = $site->Occupancy();
            ID        = sprintf("%s%04d", $atomname, int($occupancy*1000.0+1e-3));

            r = $this->GetInterAtomicDistance($xc, $yc, $zc, $x+$nx, $y+$ny, $z+$nz);
            next if($r > $Rmax);

            my %h = (
                ID          => $ID,
                'name'      => $atomname,
                'x'         => $x+$nx,
                'y'         => $y+$ny,
                'z'         => $z+$nz,
                'occupancy' => $occupancy,
                'r'         => $r,
                );

            pRArray{$ID} = [] if(!$pRArray{$ID});
            p = $pRArray{$ID};
            push(@$p, \%h);
        }
    }
    }
    }

    my ($dx, $dy, $dz) = (0, 0, 0);
    c = 0;
    my @F = (1, 1, 1);
    foreach my $key (sort keys %pRArray) {
        p = $pRArray{$key};
        @$p = sort { $a->{r} <=> $b->{r} } @$p;

        my ($i0, $Rmin, $Rth, $n1stCoord);
        if($p->[0]{r} < 3.0*$kRth) {
            Rmin = $p->[1]{r}-$p->[0]{r};
            i0 = 2;
        }
        else {
            Rmin = $p->[0]{r};
            i0 = 1;
        }
        Rth  = $Rmin * $kRth;
        my ($xc1, $yc1, $zc1) = (0.0, 0.0, 0.0);
        for(my $i = 0 ; $i < @$p ; $i++) {
            if($i >= $i0 and $p->[$i]{r} - $p->[$i-1]{r} > $Rth) {
                n1stCoord = $i;
                last;
            }
#            ($xc1, $yc1, $zc1) += ($p->[$i-1]{x}, $p->[$i-1]{y}, $p->[$i-1]{z});
            xc1 += $p->[$i]{x};
            yc1 += $p->[$i]{y};
            zc1 += $p->[$i]{z};
        }

#        ($xc1, $yc1, $zc1) /= $n1stCoord;
        if($n1stCoord > 1) {
            xc1 /= $n1stCoord;
            yc1 /= $n1stCoord;
            zc1 /= $n1stCoord;
        }

        dx += $xc1;
        dy += $yc1;
        dz += $zc1;
        c++;

        printf "%7s: # of atoms in 1st coordination = %2d: Center (%8.5f, %8.5f, %8.5f)\n", 
            key, $n1stCoord, $xc1, $yc1, $zc1;

print "    Find site symmetry at ($xc1, $yc1, $zc1)\n";
        my ($ret, $pMatchedSyms) = $this->FindSiteSymmetries($xc1, $yc1, $zc1, $OverlapCriteriaRadius);#, $IsPrint);
        for(my $is = 0 ; $is < @MArrays ; $is++) {
            next if(!$pMatchedSyms->[$is]);

#            pM   = $MArrays[$is]->{Matrix};
#            Name = $MArrays[$is]->{Name};
            pF   = $MArrays[$is]->{Freedom};
            
            F[0] = 0 if(!$pF->[0]);
            F[1] = 0 if(!$pF->[1]);
            F[2] = 0 if(!$pF->[2]);
        }
print "   ** Total freedom: ($F[0], $F[1], $F[2])\n";

        if($ShowAllCoordination) {
            for(my $i = 0 ; $i < @$p ; $i++) {
                printf "   $i: (%8.5f, %8.5f, %8.5f)) r=%8.5f  occ=%f\n",
                        p->[$i]{x}, $p->[$i]{y}, $p->[$i]{z}, $p->[$i]{r}, $p->[$i]{occupancy};
            }
        }
    }
    print "\n";

    $dx /= $c;
    $dy /= $c;
    $dz /= $c;
#exit;

    return ($dx, $dy, $dz, \@F);
}

sub FindPossibleSPGs
{
    my ($this, $ShowCrystalInf, $ShowAllCoordination, $ShowFoundStructure, $ScanAll,
        FindSiteSymmetry, $xc0, $yc0, $zc0, 
        OverlapCriteriaRadius, $ShowAllCoordination2, $RCoordMax, $IsPrint) = @_;
#        OverlapCriteriaRadius, $ShowAllCoordination, $RCoordMax, $IsPrint) = @_;

    $this->ExpandCoordinates();

    SPG                = $this->GetCSpaceGroup();
    SPGName            = $SPG->SPGName();
    iSPG               = $SPG->iSPG();
    LatticeSystem      = $SPG->LatticeSystem();
    nTranslation       = $SPG->nTranslation();
    nSymmetryOperation = $SPG->nSymmetryOperation();

    my @AtomTypeList           = $this->GetCAtomTypeList();
    nAtomType              = @AtomTypeList;
    my @AtomSiteList           = $this->GetCAsymmetricAtomSiteList();
    nAsymmetricAtomSite    = @AtomSiteList;
    my @ExpandedAtomSiteList   = $this->GetCExpandedAtomSiteList();
    nTotalExpandedAtomSite = $this->nTotalExpandedAtomSite();

    my ($TSym, $pID) = $this->FindTranslations($OverlapCriteriaRadius);
    if($TSym ne '') {
        print "** Translation [$TSym] is found\n";
        if($ShowCrystalInf) {
            print "  Asymmetric sites:\n";
            c = 0;
            for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
                next if($pID->[$i]);
    
                p         = $ExpandedAtomSiteList[$i];
                label     = $p->Label();
                type      = $p->AtomName();
                my ($x,$y,$z) = $p->Position(1);
                occupancy = $p->Occupancy();
                iAtomType = $this->FindiAtomType($type);
                printf "   %3d: %4s (%2s)[#%02d]: (%8.5f, %8.5f, %8.5f) [%f]\n",
                            c, $label, $type, $iAtomType, $x, $y, $z, $occupancy;
                c++;
            }
        }
        print "\n";
    }

    $TSym = '' if($ScanAll);

    my ($xc, $yc, $zc, $pFreedom) = $this->FindOrigin($xc0, $yc0, $zc0, $OverlapCriteriaRadius,
            ShowAllCoordination, $RCoordMax);
    $xc = 0 if(!$pFreedom->[0]);
    $yc = 0 if(!$pFreedom->[1]);
    $zc = 0 if(!$pFreedom->[2]);
    print "   Translation vector: (", -$xc, ", ", -$yc, ", ", -$zc, ")\n";

    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        p              = $ExpandedAtomSiteList[$i];
        atomname       = $p->AtomNameOnly();
        my ($x, $y, $z)    = $p->Position();
        my ($x1, $y1, $z1) = ($x-$xc, $y-$yc, $z-$zc);
        p->SetPosition($x1, $y1, $z1);
        printf "  Atom[%3d]: %2s: (%8.5f, %8.5f, %8.5f) to (%8.5f,%8.5f,%8.5f)\n",
                    i, $atomname, $x, $y, $z,$x1,$y1,$z1 if($ShowCrystalInf);
    }
    $this->SetCAsymmetricAtomSiteList(@ExpandedAtomSiteList);
#    @ExpandedAtomSiteList = $this->GetCExpandedAtomSiteList();

    my ($MostProbableSPG, $MostProbableiSPG, $MostProbableiSet, $MaxnSym) = (undef, -1, -1, 0);
    my @PossibleSPGs;
    c = 0;
    for(my $iSPG = 230 ; $iSPG >= 1 ; $iSPG--) {
        for(my $iSet = 1 ; ; $iSet++) {
            my ($SPG2, $IsMatched, $mess, $pSiteId, $pAsymSites) 
                    = $this->CheckSPG($TSym, $pID, $iSPG, $iSet, $OverlapCriteriaRadius, 0);
            last if(!defined $SPG2);

            if($IsMatched) {
                nASSite            = @$pAsymSites;
                SPGName2           = $SPG2->SPGName();
                LatticeSystem2     = $SPG2->LatticeSystem();
                nTranslation       = $SPG2->nTranslation();
                nSymmetryOperation = $SPG2->nSymmetryOperation();
                nTotalSymmetry     = $nTranslation * $nSymmetryOperation;
                if($MaxnSym < $nTotalSymmetry) {
                    MaxnSym = $nTotalSymmetry;
                    MostProbableSPG  = $SPG2;
                    MostProbableiSPG = $iSPG;
                    MostProbableiSet = $iSet;
                }
                PossibleSPGs[$c] = {
                    pSPG               => $SPG2,
                    iSPG               => $iSPG,
                    iSet               => $iSet,
                    SPGName            => $SPGName2,
                    LatticeSystem      => $LatticeSystem2,
                    nTranslation       => $nTranslation,
                    nSymmetryOperation => $nSymmetryOperation,
                    nTotalSymmetry     => $nTotalSymmetry,
                    pAsymSites         => $pAsymSites,
                    nASSite            => $nASSite,
                    };
                c++;
            }
            else {
#                print "SPG $SPGName2 (#$iSPG-$iSet LS=$LatticeSystem2) is not matched [$mess]\n";
            }
        }
    }

    SPGName2           = $MostProbableSPG->SPGName();
    LatticeSystem2     = $MostProbableSPG->LatticeSystem();
    $nTranslation          = $MostProbableSPG->nTranslation();
    $nSymmetryOperation    = $MostProbableSPG->nSymmetryOperation();
    nTotalSymmetry     = $nTranslation * $nSymmetryOperation;
    my %MostPossibleSPG = (
        iSPG               => $MostProbableiSPG,
        iSet               => $MostProbableiSet,
        SPGName            => $SPGName2,
        LatticeSystem      => $LatticeSystem2,
        nTranslation       => $nTranslation,
        nSymmetryOperation => $nSymmetryOperation,
        nTotalSymmetry     => $nTotalSymmetry
        );

    return (\%MostPossibleSPG, \@PossibleSPGs);
}

sub ExpandToSuperLattice
{
    my ($this,$nx,$ny,$nz,$crystal) = @_;

my $LF="<br>\n";
print "<H3>Expand to SuperLattice in Crystal.pm.</H3>\n";

    $nx = int($nx + 0.01);
    $ny = int($ny + 0.01);
    $nz = int($nz + 0.01);
    $nx = 1 if($nx < 1);
    $ny = 1 if($ny < 1);
    $nz = 1 if($nz < 1);
    nMult = $nx*$ny*$nz;

    $this->SetCrystalName($crystal->CrystalName());
    $this->SetCAtomTypeList($crystal->GetCAtomTypeList());
    $this->SetnAtomType($crystal->nAtomType());
    $this->SetCAsymmetricAtomSiteList($crystal->GetCAsymmetricAtomSiteList());
    $this->SetnAsymmetricAtomSite($crystal->nAsymmetricAtomSite());
    $this->SetDensity($crystal->Density());
    my ($a,$b,$c,$alpha,$beta,$gamma) = $crystal->LatticeParameters();
    $this->SetLatticeParameters($a*$nx, $b*$ny, $c*$nz, $alpha, $beta, $gamma);

    my @ExpandedAtomSiteList   = $crystal->GetCExpandedAtomSiteList();
    nTotalExpandedAtomSite = $crystal->nTotalExpandedAtomSite();

#超格子へ展開する
    my @AtomSiteList;
    count = 0;
    for(my $i = 0 ; $i < $nTotalExpandedAtomSite ; $i++) {
        label     = $ExpandedAtomSiteList[$i]->Label();
        type      = $ExpandedAtomSiteList[$i]->AtomName();
        my ($x,$y,$z) = $ExpandedAtomSiteList[$i]->Position(1);
        occupancy = $ExpandedAtomSiteList[$i]->Occupancy();
        id        = $ExpandedAtomSiteList[$i]->IdAsymmetricAtomSite();
        
        for(my $ix = 0 ; $ix < $nx ; $ix++) {
        for(my $iy = 0 ; $iy < $ny ; $iy++) {
        for(my $iz = 0 ; $iz < $nz ; $iz++) {
            x0 = ($x + $ix) / $nx;
            y0 = ($y + $iy) / $ny;
            z0 = ($z + $iz) / $nz;

            AtomSiteList[$count] = new AtomSite();
            AtomSiteList[$count]->SetLabel($label);
            AtomSiteList[$count]->SetAtomName($type);
            AtomSiteList[$count]->SetPosition($x0, $y0, $z0);
            AtomSiteList[$count]->SetOccupancy($occupancy);
            AtomSiteList[$count]->SetIdAsymmetricAtomSite($id);

            count++;
        }
        }
        }
    }
#print "count: $count\n";
    $this->{"nTotalExpandedAtomSite"}  = $count;
    $this->{"RefExpandedAtomSiteList"} = \@AtomSiteList;

#多重度を超格子の値に更新
    my @nMultiplicityExpandedAtomSiteList
         = $crystal->GetCnMultiplicityExpandedAtomSiteList();
    for(my $i = 0 ; $i < @nMultiplicityExpandedAtomSiteList ; $i++) {
        nMultiplicityExpandedAtomSiteList[$i] *= $nMult;
    }
    $this->{"RefnExpandedAtomSite"} = \@nMultiplicityExpandedAtomSiteList;

#多重度を、全原子に設定
    for(my $i = 0 ; $i < @AtomSiteList ; $i++) {
        atom = $AtomSiteList[$i];
        ia   = $atom->Id() - 1;
        atom->SetMultiplicity($nMultiplicityExpandedAtomSiteList[$ia]);
        this->SetIsSelected($i, 0);
    }

#IDをもとに部分占有原子を整理
#Occupancyに合わせて表示する原子数を選択する => IsSelectedに選択された原子を設定する。

#同じ位置にあるサイトの数
    my @nIdenticalSite;
#そのサイトを実際に占める原子の数
    my %nOccupiedAtom;
    my %iIdenticalPartialAtomWithSameId;
#ID=$idのサイトで、$j番目のサイト占める原子の名前　$AtomTypeInPartialAtom{"[$id][$j]"}
    my %AtomTypeInPartialAtom;
#ID=$idのサイトで、$j番目のサイトを占める原子のID　$AtomIdInPartialAtom{"[$id][$j]"}
    my %AtomIdInPartialAtom;
#ID=$idのサイトで、$j番目のサイトを占める原子のIndex　$AtomIndexInPartialAtom{"[$id][$j]"}
    my %AtomIndexInPartialAtom;
#多重度
    my @nSameIdSite;
    my %SameIdSiteIndex;


#同じIDをもつサイトのindexを$SameIdSiteIndex{"[$idm1][$sid]"} に設定する
    for(my $i = 0 ; $i < @ExpandedAtomSiteList ; $i++){
        atom      = $ExpandedAtomSiteList[$i];
        type      = $atom->AtomName();
        my ($x,$y,$z) = $atom->Position(1);
        occ       = $atom->Occupancy();
        mult      = $atom->Multiplicity();
        id        = $atom->Id();
        idm1      = $id-1;
        nSameIdSite[$idm1]++;
        sid = $nSameIdSite[$idm1]-1;
        SameIdSiteIndex{"[$idm1][$sid]"} = $i;
#printf "#++%4d[%4s]: (%7.4f, %7.4f, %7.4f) [%6.4f] (mult=%3d: id=%2d)$LF",
#    $i+1, $type, $x, $y, $z, $occ, $mult, $id;
    }

print "<b>#Check number of atoms occupying the sites.</b>$LF";
    my @IsPrintedById;
    my @IsPrinted;
    for(my $i = 0 ; $i < @ExpandedAtomSiteList ; $i++) {
        PrintCount = 0;
        atom = $ExpandedAtomSiteList[$i];
        type = $atom->AtomName();
        my ($x,$y,$z) = $atom->Position(1);
        occ  = $atom->Occupancy();
        id   = $atom->Id();
        mult = $atom->Multiplicity();
        idm1 = $id-1;
#該当する等価な全てサイトのうち、何個の原子が存在するか
        nOccupiedAtom{"[$i][$id]"} = int($mult * $occ + 0.01);
        nc = $nOccupiedAtom{"[$i][$id]"};
        next if($IsPrintedById[$id]);
        next if($occ >= 0.999999);

#print "+++A(i:$i): $type: occ=$occ: nMult=$mult: nAtom=$nc (id=$id)$LF";

#同じ位置にある原子のIDのリストを作る
        iIdenticalPartialAtomWithSameId{"[$i][0]"} = $i;

        IsPrinted[$i] = 1;
        IsPrintedById[$id] = 1;

#同じ位置にあるサイトの数を$cにいれる
        c = 1;
        for(my $j = $i+1 ; $j < @ExpandedAtomSiteList ; $j++) {
            next if($IsPrinted[$j]);
            atom2 = $ExpandedAtomSiteList[$j];
            type2 = $atom2->AtomName();;
            my ($x2,$y2,$z2) = $atom2->Position(1);
            occ2  = $atom2->Occupancy();
            id2   = $atom2->Id();
            mult2 = $atom2->Multiplicity();

            next if($occ2 >= 0.999999);
            dis = $this->GetNearestInterAtomicDistance(
                    x, $y, $z, $x2, $y2, $z2);
            next if($dis > 0.001);

            iIdenticalPartialAtomWithSameId{"[$i][$c]"} = $j;

#$i番目のサイトと同じ位置にあるID=$id2のサイト、実際にいくつの原子があるか
            nOccupiedAtom{"[$i][$id2]"} = int($mult2 * $occ2 + 0.01);

            IsPrinted[$j] = 1;
            IsPrintedById[$id2] = 1;
            c++;

my $nc2 = $nOccupiedAtom{"[$i][$id2]"};
#print "+++***(j:$j)$type2: occ=$occ2: nMult=$mult2: nAtom=$nc2 (id=$id2)$LF";
        }

#$i番と同じ位置にあるサイトサイトの数を$cにいれる
#$iは、0から始まる（ID-1に等しい）
        nIdenticalSite[$i] = $c;
    }

#結果の表示
    for(my $id0 = 0 ; $id0 < @ExpandedAtomSiteList ; $id0++) {
#$nIdenticalSiteは部分共有原子の$iのみで定義される。
#同じ位置にある原子の値は0なので、この条件文でスキップできる
        next if(!defined $nIdenticalSite[$id0]or $nIdenticalSite[$id0] <= 0);

        i  = $iIdenticalPartialAtomWithSameId{"[$id0][0]"};
        atom = $ExpandedAtomSiteList[$i];
        my ($x0,$y0,$z0) = $atom->Position(1);
        Id           = $atom->Id($i);
        occupancy    = $atom->Occupancy($i);
        mult         = $atom->Multiplicity($i);
        Idm1 = $Id-1;

#$nIdenticalSiteは0から始まる配列
print "<b>ID=</b>$Id: nIdenticalSite = ", $nIdenticalSite[$id0], "  at ($x0, $y0, $z0)$LF";
print "+++ i=";
        for(my $sn = 0 ; $sn < $nIdenticalSite[$id0] ; $sn++) {
            j = $iIdenticalPartialAtomWithSameId{"[$id0][$sn]"};
print "$j ";
        }
print "$LF";

        for(my $sn = 0 ; $sn < $nIdenticalSite[$id0] ; $sn++) {
            j = $iIdenticalPartialAtomWithSameId{"[$i][$sn]"};
            next if(!defined($j) or $j eq '');
            atom  = $ExpandedAtomSiteList[$j];
            Id2          = $atom->Id();
            Id2m1        = $Id2-1;
            type2        = $atom->AtomName();
            my ($x2,$y2,$z2) = $atom->Position(1);
            occ2         = $atom->Occupancy();
            mult2        = $atom->Multiplicity();
            nAtom2 = $nOccupiedAtom{"[$i][$Id2]"};

print "+++ i=$j(Id=$Id2): $type2 ($x2, $y2, $z2)"
    . " [occ=$occ2] [nMult=$mult2] [N=$nAtom2]$LF";
print "+++*** SameIdSiteIndex";
            for(my $j = 0 ; $j < $mult ; $j++) {
#                idx = $this->SameIdSiteIndex($Id2m1, $j);
                idx = $SameIdSiteIndex{"[$Id2m1][$j]"};
printf "%3d ", $idx;
            }
print "$LF";
        }


#$SameIdSiteIndex{"[$Id2][$j]"}から、実際に割り当てる原子を決める
        nSiteWithSameId0 = $atom->Multiplicity();
        ResidualNum      = $nSiteWithSameId0;
        nIdenticalSite0  = $nIdenticalSite[$i];
#print "nSiteWithSameId0: $nSiteWithSameId0, nIdenticalSite0: $nIdenticalSite0$LF";

        for(my $sn = 0 ; $sn < $nIdenticalSite0 ; $sn++) {
            jatom = $iIdenticalPartialAtomWithSameId{"[$i][$sn]"};
#print "jatom=$jatom$LF";
            atom0 = $ExpandedAtomSiteList[$jatom];
            Id0   = $atom0->Id();
            type  = $atom0->AtomName();
            nAtom = $nOccupiedAtom{"[$i][$Id0]"};
#print "j=$jatom: Id=$Id0: $type: nOccupiedAtom=$nAtom$LF";
            for(my $is = 0 ; $is < $nAtom ; $is++) {
                r = int(rand($ResidualNum));
                for(my $j = 0 ; $j < $nSiteWithSameId0 ; $j++) {
                    next if(defined($AtomTypeInPartialAtom{"[$Id][$j]"}));
                    r--;
                    if($r < 0) {
#ID=$idの原子と同一のサイトで、$j番目のサイトをしめる原子 $AtomTypeInPartialAtom{"[$id][$j]"}
                        AtomTypeInPartialAtom{"[$Id][$j]"} = $type;
                        AtomIdInPartialAtom{"[$Id][$j]"}   = $Id0;
                        Id0m1 = $Id0 - 1;
                        Index = $SameIdSiteIndex{"[$Id0m1][$j]"}; 
#表示する原子を設定する

                        this->SetIsSelected($Index, 1);
                        AtomIndexInPartialAtom{"[$Id][$j]"} = $Index;
#print "Id0=$Id0, sn=$j, Idx=$Index$LF";
                        ResidualNum--;
                        last;
                    }
                }
            }
        }
    }

if(0) {
print "<b>Summary:</b>$LF";
#    my @ExpandedAtomSiteList = $this->GetCExpandedAtomSiteList();
    for(my $i = 0 ; $i < @ExpandedAtomSiteList ; $i++) {
        atom         = $ExpandedAtomSiteList[$i];
        name         = $atom->AtomName();
        occupancy    = $atom->Occupancy();
#表示する原子を設定する
        this->SetIsSelected($i, 1) if($occupancy >= 1);
        next if(!$this->IsSelected($i));

        my ($x,$y,$z)    = $atom->Position(1);
        Id           = $atom->Id();
print "+++$i: $name (id=$Id): ($x, $y, $z) [$occupancy]$LF";
    }
}

print "<H3>Expand to Superlattice in Crystal.pm: End</H3>\n";
print "$LF";

    return $count;
}

sub SetOutputMode
{
    my ($this, $mode) = @_;
    return $this->{'OutputMode'} = $mode;
}

sub OutputMode
{
    my ($this) = @_;
    return $this->{'OutputMode'};
}

sub SPGNameByOutputMode
{
    my ($this) = @_;

    if($this->{OutputMode} eq 'expanded' or
       $this->{OutputMode} eq 'choose' ) {
        return "P 1";
    }
    return $this->SPGName();
}

sub iSPGByOutputMode
{
    my ($this) = @_;
    if($this->{'OutputMode'} eq 'expanded' or
       $this->{'OutputMode'} eq 'choose' ) {
        return 1;
    }
    return $this->iSPG();
}

sub iSetByOutputMode
{
    my ($this) = @_;
    if($this->{'OutputMode'} eq 'expanded' or
       $this->{'OutputMode'} eq 'choose' ) {
        return 1;
    }
    return $this->iSet();
}

sub FractionalToCartesianByOutputMode
{
    my ($this, $x,$y,$z) = @_;

    my ($a11, $a12, $a13, $a21, $a22, $a23, $a31, $a32, $a33);
    if($this->{'OutputMode'} eq 'expanded' or
       $this->{'OutputMode'} eq 'choose' ) {
        SuperLattice = $this->{'CSuperLattice'};
        a11 = $SuperLattice->{"a11"};
        a12 = $SuperLattice->{"a12"};
        a13 = $SuperLattice->{"a13"};
        a21 = $SuperLattice->{"a21"};
        a22 = $SuperLattice->{"a22"};
        a23 = $SuperLattice->{"a23"};
        a31 = $SuperLattice->{"a31"};
        a32 = $SuperLattice->{"a32"};
        a33 = $SuperLattice->{"a33"};

    }
    else {
        a11 = $this->{"a11"};
        a12 = $this->{"a12"};
        a13 = $this->{"a13"};
        a21 = $this->{"a21"};
        a22 = $this->{"a22"};
        a23 = $this->{"a23"};
        a31 = $this->{"a31"};
        a32 = $this->{"a32"};
        a33 = $this->{"a33"};
    }
    xc = $a11*$x + $a21*$y + $a31*$z;
    yc = $a12*$x + $a22*$y + $a32*$z;
    zc = $a13*$x + $a23*$y + $a33*$z;
    return ($xc,$yc,$zc);
}


sub LatticeParametersByOutputMode
{
    my ($this, $UseAtomicUnit) = @_;
    if($this->{'OutputMode'} eq 'expanded' or
       $this->{'OutputMode'} eq 'choose' ) {
        SuperLattice = $this->{'CSuperLattice'};
        return $SuperLattice->LatticeParameters($UseAtomicUnit)
            if($SuperLattice);
        return $this->LatticeParameters($UseAtomicUnit);
    }
    return $this->LatticeParameters($UseAtomicUnit);
}

sub LatticeVectorsByOutputMode
{
    my ($this, $UseAtomicUnit) = @_;
    k = ($UseAtomicUnit)? 1.0 / 0.529177 : 1.0;
#my $SPG     = $this->GetCSpaceGroup();
#my $SPGName = $SPG->SPGName();
#print "SPG in Crystal.pm:[$SPGName]\n";
#exit;
    if($this->{'OutputMode'} eq 'expanded' or
       $this->{'OutputMode'} eq 'choose' ) {
        SuperLattice = $this->{'CSuperLattice'};
        return $SuperLattice->LatticeVectors($UseAtomicUnit)
            if($SuperLattice);
        return $this->LatticeVectors($UseAtomicUnit);
    }
    return $this->LatticeVectors($UseAtomicUnit);
}

sub GetCExpandedAtomSiteListByOutputMode
{
    my ($this,$UseAtomicUnit) = @_;
    if(defined $this->{'OutputMode'} and $this->{'OutputMode'} eq 'expanded') {
        SuperLattice = $this->{'CSuperLattice'};
        return $SuperLattice->GetCExpandedAtomSiteList()
            if($SuperLattice);
        return $this->GetCExpandedAtomSiteList();
    }
    elsif(defined $this->{'OutputMode'} and $this->{'OutputMode'} eq 'choose' ) {
        SuperLattice = $this->{'CSuperLattice'};
        if($SuperLattice) {
            my @slist = $SuperLattice->GetCExpandedAtomSiteList();
            my @tlist;
            for(my $i = 0 ; $i < @slist ; $i++) {
                atom = $slist[$i];
                next if($atom->Occupancy() < 1.0 and
                    !$atom->IsSelected() );
                push(@tlist, $atom);
            }
            return @tlist;
        }
        return $this->GetCExpandedAtomSiteList();
    }
    return $this->GetCAsymmetricAtomSiteList();
}

sub CreateSuperLattice
{
    my ($this, $nx, $ny, $nz) = @_;

    SuperLattice = new Crystal();
    $SuperLattice->ExpandToSuperLattice($nx,$ny,$nz,$this);
    $this->{'CSuperLattice'} = $SuperLattice;
    return 1;
}

sub GetCSuperLattice
{
    my ($this) = @_;
    return $this->{'CSuperLattice'};
}
"""