import sys
import os
from numpy import sin, cos, tan, arcsin, arccos, arctan, exp, log, sqrt
import numpy as np
from numpy import linalg as la
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt


from tklib.tkfile import tkFile
import tklib.tkutils
from tklib.tkcrystal.tkcif import tkCIF
from tklib.tkcrystal.tkcrystal import tkCrystal
from tklib.tkcrystal.tkatomtype import tkAtomType
from tkcrystalbase import *




"""
Calculate Bragg angles
  Requirement: tkcrystalbase.py
"""


infile = None


# Lattice parameters (angstrom and degree)
#lattice_parameters = [ 5.62, 5.62, 5.62, 60.0, 60.0, 60.0]
lattice_parameters = [ 5.62, 5.62, 5.62, 90.0, 90.0, 90.0]

# Site information (atom name, site label, atomic number, atomic mass, charge, radius, color, position)
sites = [
         ['Na', 'Na1', 11, 22.98997, +1.0, 0.7, 'red',  np.array([0.0, 0.0, 0.0])]
        ,['Na', 'Na2', 11, 22.98997, +1.0, 0.7, 'red',  np.array([0.0, 0.5, 0.5])]
        ,['Na', 'Na3', 11, 22.98997, +1.0, 0.7, 'red',  np.array([0.5, 0.0, 0.5])]
        ,['Na', 'Na4', 11, 22.98997, +1.0, 0.7, 'red',  np.array([0.5, 0.5, 0.0])]
        ,['Cl', 'Cl1', 17, 35.4527,  -1.0, 1.4, 'blue', np.array([0.5, 0.0, 0.0])]
        ,['Cl', 'Cl2', 17, 35.4527,  -1.0, 1.4, 'blue', np.array([0.5, 0.5, 0.5])]
        ,['Cl', 'Cl3', 17, 35.4527,  -1.0, 1.4, 'blue', np.array([0.0, 0.0, 0.5])]
        ,['Cl', 'Cl4', 17, 35.4527,  -1.0, 1.4, 'blue', np.array([0.0, 0.5, 0.0])]
        ]

# X-ray wavelength
wl = 1.5405         # angstrom

# G min to remove the origin of reciprocal space
Gmin = 1.0e-5

# 2Theta max
Q2max = 150.0       # degree in 2Theta


def main():
    global lattice_parameters, sites
    
    if infile is not None:
        print("")
        print("=============== CIF file read test ============")
        print("infile: {}".format(infile))

        print("")
        print("Read [{}]".format(infile))
        cif = tkCIF()
        cif.debug = False
        cifdata = cif.ReadCIF(infile, find_valid_structure = True)
        cif.Close()
        if not cifdata:
            terminate("Error: Could not get cifdat from infile [{}]".format(infile))

#        cifdata.print()
        cry = cifdata.GetCrystal()
        cry.PrintInf()

        lattice_parameters = cry.LatticeParameters()
        aij = cry.LatticeVectors()
        gij, Rgij = cry.Metrics()
        volume = cry.Volume()

        Raij = cry.ReciprocalLatticeVectors()
        Rlatt = cry.ReciprocalLatticeParameters()
        Rvolume = cry.ReciprocalVolume()

        print("")
        print("Lattice parameters:", lattice_parameters)
        print("Lattice vectors:")
        print("  ax: ({:10.4g}, {:10.4g}, {:10.4g}) A".format(aij[0][0], aij[0][1], aij[0][2]))
        print("  ay: ({:10.4g}, {:10.4g}, {:10.4g}) A".format(aij[1][0], aij[1][1], aij[1][2]))
        print("  az: ({:10.4g}, {:10.4g}, {:10.4g}) A".format(aij[2][0], aij[2][1], aij[2][2]))
        print("Metric tensor:")
        print("  gij: ({:10.4g}, {:10.4g}, {:10.4g}) A".format(*gij[0]))
        print("       ({:10.4g}, {:10.4g}, {:10.4g}) A".format(*gij[1]))
        print("       ({:10.4g}, {:10.4g}, {:10.4g}) A".format(*gij[2]))
        print("Unit cell volume: {:12.4g} A^3".format(volume))

        print("")
        print("Reciprocal lattice parameters:", Rlatt)
        print("Reciprocal lattice vectors:")
        print("  Rax: ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Raij[0]))
        print("  Ray: ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Raij[1]))
        print("  Raz: ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Raij[2]))
        print("Reciprocal lattice metric tensor:")
        print("  Rgij: ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Rgij[0]))
        print("        ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Rgij[1]))
        print("        ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Rgij[2]))
        print("Reciprocal unit cell volume: {:12.4g} A^-3".format(Rvolume))

        sites = []
        AtomTypes = cry.AtomTypeList()
        AtomSites = cry.ExpandedAtomSiteList()
        allsites = []
        for atom in AtomSites:
            label  = atom.Label()
            name   = atom.AtomNameOnly()
            z      = 1 #atom.AtomicNumber()
            q      = atom.Charge()
            pos    = atom.Position()
            xc, yc, zc = cry.FractionalToCartesian(*pos)
            if q >= 0.0:
                r = 0.7
                color = 'red'
                M = 1.0
            else:
                r = 1.4
                color = 'blue'
                M = 1.0
            
            sites.append([name, label, z, M, q, r, color, pos])
    else:
        print("")
        print("Lattice parameters:", lattice_parameters)
        aij = cal_lattice_vectors(lattice_parameters)
        print("Lattice vectors:")
        print("  ax: ({:10.4g}, {:10.4g}, {:10.4g}) A".format(aij[0][0], aij[0][1], aij[0][2]))
        print("  ay: ({:10.4g}, {:10.4g}, {:10.4g}) A".format(aij[1][0], aij[1][1], aij[1][2]))
        print("  az: ({:10.4g}, {:10.4g}, {:10.4g}) A".format(aij[2][0], aij[2][1], aij[2][2]))
        inf = cal_metrics(lattice_parameters)
        gij = inf['gij']
        print("Metric tensor:")
        print("  gij: ({:10.4g}, {:10.4g}, {:10.4g}) A".format(*gij[0]))
        print("       ({:10.4g}, {:10.4g}, {:10.4g}) A".format(*gij[1]))
        print("       ({:10.4g}, {:10.4g}, {:10.4g}) A".format(*gij[2]))
        volume = cal_volume(aij)
        print("Volume: {:12.4g} A^3".format(volume))

        print("")
        print("Unit cell volume: {:12.4g} A^3".format(volume))
        Raij  = cal_reciprocal_lattice_vectors(aij)
        Rlatt = cal_reciprocal_lattice_parameters(Raij)
        Rinf  = cal_metrics(Rlatt)
        Rgij  = Rinf['gij']
        print("Reciprocal lattice parameters:", Rlatt)
        print("Reciprocal lattice vectors:")
        print("  Rax: ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Raij[0]))
        print("  Ray: ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Raij[1]))
        print("  Raz: ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Raij[2]))
        print("Reciprocal lattice metric tensor:")
        print("  Rgij: ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Rgij[0]))
        print("        ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Rgij[1]))
        print("        ({:10.4g}, {:10.4g}, {:10.4g}) A^-1".format(*Rgij[2]))
        Rvolume = cal_volume(Raij)
        print("Reciprocal unit cell volume: {:12.4g} A^-3".format(Rvolume))

    dmin = wl / 2.0 / sin(0.5 * Q2max * torad)
    hmax = int(lattice_parameters[0] / dmin)
    kmax = int(lattice_parameters[1] / dmin)
    lmax = int(lattice_parameters[2] / dmin)

    print("")
    print("hkl range:", hmax, kmax, lmax)

# Calculate diffraction angles and store them in qlist list variable
    org = np.array([0.0, 0.0, 0.0])
    qlist = []
    for l in range(-lmax, lmax+1):
      for k in range(-kmax, kmax+1):
        for h in range(-hmax, hmax+1):
# Calculate distance in reciprocal space between (0, 0, 0) and (h, k, l)
            G = distance(org, np.array([h, k, l]), Rgij)
            if G < Gmin:
                continue

# Calculate lattice space from G
            d = 1.0 / G

            sinQ = wl / 2.0 / d
            if sinQ >= 1.0:
                continue
            
# Calculate diffraction angle 2Theta
            Q2 = 2.0 * todeg * arcsin(sinQ)
            if Q2 > Q2max:
                continue

            qlist.append([Q2, d, h, k, l])
#            print("  2Q={:12.4g}  d={:12.6g}  ({:3d} {:3d} {:3d})".format(Q2, d, h, k, l))

# Sort rlist by 2Theta (x[0] priority)
    qlist.sort(key = lambda x: (x[0], x[2], x[3], x[4]))

    print("")
    print("Diffraction angle, d, h, k, l:")
    for qinf in qlist:
        Q2, d, h, k, l = qinf
        print("  2Q={:12.4g}  d={:12.6g}  ({:3d} {:3d} {:3d})".format(Q2, d, h, k, l))

    print("")
    exit()


if __name__ == '__main__':
    main()
